from django.apps import apps
from django.conf import settings
from django.contrib import admin
from django.contrib.auth import views as auth_views
from django.urls import include, path
from django.views.decorators.csrf import csrf_exempt
from django.views.i18n import JavaScriptCatalog

import calendarweek.django
import debug_toolbar
from ckeditor_uploader import views as ckeditor_uploader_views
from django_js_reverse.views import urls_js
from health_check.urls import urlpatterns as health_urls
from oauth2_provider.views import ConnectDiscoveryInfoView
from rules.contrib.views import permission_required
from two_factor.urls import urlpatterns as tf_urls

from . import views

urlpatterns = [
    path("", include("django_prometheus.urls")),
    path(settings.MEDIA_URL.removeprefix("/"), include("titofisto.urls")),
    path("manifest.json", views.ManifestView.as_view(), name="manifest"),
    path("serviceworker.js", views.ServiceWorkerView.as_view(), name="service_worker"),
    path("vue_dummy/", views.vue_dummy, name="vue_dummy"),
    path("offline/", views.OfflineView.as_view(), name="offline"),
    path("about/", views.about, name="about_aleksis"),
    path("accounts/signup/", views.AccountRegisterView.as_view(), name="account_signup"),
    path("accounts/logout/", auth_views.LogoutView.as_view(), name="logout"),
    path(
        "accounts/password/change/",
        views.CustomPasswordChangeView.as_view(),
        name="account_change_password",
    ),
    path(
        "accounts/password/reset/",
        views.CustomPasswordResetView.as_view(),
        name="account_reset_password",
    ),
    path("accounts/", include("allauth.urls")),
    path("invitations/send-invite", views.InvitePerson.as_view(), name="invite_person"),
    path(
        "invitations/code/enter", views.EnterInvitationCode.as_view(), name="enter_invitation_code"
    ),
    path(
        "invitations/code/generate",
        views.GenerateInvitationCode.as_view(),
        name="generate_invitation_code",
    ),
    path("invitations/disabled", views.InviteDisabledView.as_view(), name="invite_disabled"),
    path("invitations/", include("invitations.urls")),
    path(
        "accounts/social/connections/<int:pk>/delete",
        views.SocialAccountDeleteView.as_view(),
        name="delete_social_account_by_pk",
    ),
    path("admin/", admin.site.urls),
    path("admin/uwsgi/", include("django_uwsgi.urls")),
    path("data_management/", views.data_management, name="data_management"),
    path("status/", views.SystemStatus.as_view(), name="system_status"),
    path("account/login/", views.LoginView.as_view(), name="login"),
    path("", include(tf_urls)),
    path("celery_progress/<str:task_id>/", views.CeleryProgressView.as_view(), name="task_status"),
    path("accounts/logout/", auth_views.LogoutView.as_view(), name="logout"),
    path("school_terms/", views.SchoolTermListView.as_view(), name="school_terms"),
    path("school_terms/create/", views.SchoolTermCreateView.as_view(), name="create_school_term"),
    path("school_terms/<int:pk>/", views.SchoolTermEditView.as_view(), name="edit_school_term"),
    path("persons", views.persons, name="persons"),
    path("person/", views.person, name="person"),
    path("person/create/", views.CreatePersonView.as_view(), name="create_person"),
    path("person/<int:id_>/", views.person, name="person_by_id"),
    path("person/<int:pk>/edit/", views.EditPersonView.as_view(), name="edit_person_by_id"),
    path("person/<int:id_>/delete/", views.delete_person, name="delete_person_by_id"),
    path("person/<int:pk>/invite/", views.InvitePersonByID.as_view(), name="invite_person_by_id"),
    path("groups", views.groups, name="groups"),
    path("groups/additional_fields", views.additional_fields, name="additional_fields"),
    path("groups/child_groups/", views.groups_child_groups, name="groups_child_groups"),
    path(
        "groups/additional_field/<int:id_>/edit",
        views.edit_additional_field,
        name="edit_additional_field_by_id",
    ),
    path(
        "groups/additional_field/create",
        views.edit_additional_field,
        name="create_additional_field",
    ),
    path(
        "groups/additional_field/<int:id_>/delete",
        views.delete_additional_field,
        name="delete_additional_field_by_id",
    ),
    path("group/create", views.edit_group, name="create_group"),
    path("group/<int:id_>", views.group, name="group_by_id"),
    path("group/<int:id_>/edit", views.edit_group, name="edit_group_by_id"),
    path("group/<int:id_>/delete", views.delete_group, name="delete_group_by_id"),
    path("", views.index, name="index"),
    path("notifications/", views.NotificationsListView.as_view(), name="notifications"),
    path("dashboard/edit/", views.EditDashboardView.as_view(), name="edit_dashboard"),
    path("groups/group_type/create", views.edit_group_type, name="create_group_type"),
    path(
        "groups/group_type/<int:id_>/delete",
        views.delete_group_type,
        name="delete_group_type_by_id",
    ),
    path("groups/group_type/<int:id_>/edit", views.edit_group_type, name="edit_group_type_by_id"),
    path("groups/group_types", views.group_types, name="group_types"),
    path("announcements/", views.announcements, name="announcements"),
    path("announcement/create/", views.announcement_form, name="add_announcement"),
    path("announcement/edit/<int:id_>/", views.announcement_form, name="edit_announcement"),
    path("announcement/delete/<int:id_>/", views.delete_announcement, name="delete_announcement"),
    path("search/searchbar/", views.searchbar_snippets, name="searchbar_snippets"),
    path("search/", views.PermissionSearchView.as_view(), name="haystack_search"),
    path("maintenance-mode/", include("maintenance_mode.urls")),
    path("impersonate/", include("impersonate.urls")),
    path(
        ".well-known/openid-configuration",
        ConnectDiscoveryInfoView.as_view(),
        name="oidc_configuration",
    ),
    path("oauth/applications/", views.OAuth2ListView.as_view(), name="oauth2_applications"),
    path(
        "oauth/applications/register/",
        views.OAuth2RegisterView.as_view(),
        name="register_oauth_application",
    ),
    path(
        "oauth/applications/<int:pk>/", views.OAuth2DetailView.as_view(), name="oauth2_application"
    ),
    path(
        "oauth/applications/<int:pk>/delete/",
        views.OAuth2DeleteView.as_view(),
        name="delete_oauth2_application",
    ),
    path(
        "oauth/applications/<int:pk>/edit/",
        views.OAuth2EditView.as_view(),
        name="edit_oauth2_application",
    ),
    path(
        "oauth/authorize/",
        views.CustomAuthorizationView.as_view(),
        name="oauth2_provider:authorize",
    ),
    path("oauth/", include("oauth2_provider.urls", namespace="oauth2_provider")),
    path("graphql/", csrf_exempt(views.PrivateGraphQLView.as_view(graphiql=True)), name="graphql"),
    path("__i18n__/", include("django.conf.urls.i18n")),
    path(
        "ckeditor/upload/",
        permission_required("core.ckeditor_upload_files_rule")(ckeditor_uploader_views.upload),
        name="ckeditor_upload",
    ),
    path(
        "ckeditor/browse/",
        permission_required("core.ckeditor_upload_files_rule")(ckeditor_uploader_views.browse),
        name="ckeditor_browse",
    ),
    path("select2/", include("django_select2.urls")),
    path("jsreverse.js", urls_js, name="js_reverse"),
    path("calendarweek_i18n.js", calendarweek.django.i18n_js, name="calendarweek_i18n_js"),
    path("gettext.js", JavaScriptCatalog.as_view(), name="javascript-catalog"),
    path(
        "preferences/site/", views.preferences, {"registry_name": "site"}, name="preferences_site"
    ),
    path(
        "preferences/person/",
        views.preferences,
        {"registry_name": "person"},
        name="preferences_person",
    ),
    path(
        "preferences/group/",
        views.preferences,
        {"registry_name": "group"},
        name="preferences_group",
    ),
    path(
        "preferences/site/<int:pk>/",
        views.preferences,
        {"registry_name": "site"},
        name="preferences_site",
    ),
    path(
        "preferences/person/<int:pk>/",
        views.preferences,
        {"registry_name": "person"},
        name="preferences_person",
    ),
    path(
        "preferences/group/<int:pk>/",
        views.preferences,
        {"registry_name": "group"},
        name="preferences_group",
    ),
    path(
        "preferences/site/<int:pk>/<str:section>/",
        views.preferences,
        {"registry_name": "site"},
        name="preferences_site",
    ),
    path(
        "preferences/person/<int:pk>/<str:section>/",
        views.preferences,
        {"registry_name": "person"},
        name="preferences_person",
    ),
    path(
        "preferences/group/<int:pk>/<str:section>/",
        views.preferences,
        {"registry_name": "group"},
        name="preferences_group",
    ),
    path(
        "preferences/site/<str:section>/",
        views.preferences,
        {"registry_name": "site"},
        name="preferences_site",
    ),
    path(
        "preferences/person/<str:section>/",
        views.preferences,
        {"registry_name": "person"},
        name="preferences_person",
    ),
    path(
        "preferences/group/<str:section>/",
        views.preferences,
        {"registry_name": "group"},
        name="preferences_group",
    ),
    path("health/", include(health_urls)),
    path("health/pdf/", views.TestPDFGenerationView.as_view(), name="test_pdf"),
    path(
        "data_check/",
        views.DataCheckView.as_view(),
        name="check_data",
    ),
    path(
        "data_check/run/",
        views.RunDataChecks.as_view(),
        name="data_check_run",
    ),
    path(
        "data_check/<int:pk>/<str:solve_option>/",
        views.SolveDataCheckView.as_view(),
        name="data_check_solve",
    ),
    path("dashboard_widgets/", views.DashboardWidgetListView.as_view(), name="dashboard_widgets"),
    path(
        "dashboard_widgets/<int:pk>/edit/",
        views.DashboardWidgetEditView.as_view(),
        name="edit_dashboard_widget",
    ),
    path(
        "dashboard_widgets/<int:pk>/delete/",
        views.DashboardWidgetDeleteView.as_view(),
        name="delete_dashboard_widget",
    ),
    path(
        "dashboard_widgets/<str:app>/<str:model>/new/",
        views.DashboardWidgetCreateView.as_view(),
        name="create_dashboard_widget",
    ),
    path(
        "dashboard_widgets/default/",
        views.EditDashboardView.as_view(),
        {"default": True},
        name="edit_default_dashboard",
    ),
    path(
        "permissions/global/user/",
        views.UserGlobalPermissionsListBaseView.as_view(),
        name="manage_user_global_permissions",
    ),
    path(
        "permissions/global/group/",
        views.GroupGlobalPermissionsListBaseView.as_view(),
        name="manage_group_global_permissions",
    ),
    path(
        "permissions/object/user/",
        views.UserObjectPermissionsListBaseView.as_view(),
        name="manage_user_object_permissions",
    ),
    path(
        "permissions/object/group/",
        views.GroupObjectPermissionsListBaseView.as_view(),
        name="manage_group_object_permissions",
    ),
    path(
        "permissions/global/user/<int:pk>/delete/",
        views.UserGlobalPermissionDeleteView.as_view(),
        name="delete_user_global_permission",
    ),
    path(
        "permissions/global/group/<int:pk>/delete/",
        views.GroupGlobalPermissionDeleteView.as_view(),
        name="delete_group_global_permission",
    ),
    path(
        "permissions/object/user/<int:pk>/delete/",
        views.UserObjectPermissionDeleteView.as_view(),
        name="delete_user_object_permission",
    ),
    path(
        "permissions/object/group/<int:pk>/delete/",
        views.GroupObjectPermissionDeleteView.as_view(),
        name="delete_group_object_permission",
    ),
    path(
        "permissions/assign/",
        views.SelectPermissionForAssignView.as_view(),
        name="select_permission_for_assign",
    ),
    path(
        "permissions/<int:pk>/assign/",
        views.AssignPermissionView.as_view(),
        name="assign_permission",
    ),
    path("pdfs/<int:pk>/", views.RedirectToPDFFile.as_view(), name="redirect_to_pdf_file"),
    path("ical/", views.ICalFeedListView.as_view(), name="ical_feed_list"),
    path("ical/create/", views.ICalFeedCreateView.as_view(), name="ical_feed_create"),
    path("ical/<int:pk>/edit/", views.ICalFeedEditView.as_view(), name="ical_feed_edit"),
    path("ical/<int:pk>/delete/", views.ICalFeedDeleteView.as_view(), name="ical_feed_delete"),
    path("ical/<slug:slug>.ics", views.ICalFeedView.as_view(), name="ical_feed"),
    path("__icons__/", include("dj_iconify.urls")),
]

# Use custom server error handler to get a request object in the template
handler500 = views.server_error

# Add URLs for optional features
if hasattr(settings, "TWILIO_ACCOUNT_SID"):
    from two_factor.gateways.twilio.urls import urlpatterns as tf_twilio_urls  # noqa

    urlpatterns += [path("", include(tf_twilio_urls))]

# Serve javascript-common if in development
if settings.DEBUG:
    urlpatterns.append(path("__debug__/", include(debug_toolbar.urls)))

# Automatically mount URLs from all installed AlekSIS apps
for app_config in apps.app_configs.values():
    if not app_config.name.startswith("aleksis.apps."):
        continue

    try:
        urlpatterns.append(path(f"app/{app_config.label}/", include(f"{app_config.name}.urls")))
    except ModuleNotFoundError:
        # Ignore exception as app just has no URLs
        pass  # noqa



from typing import Callable

from django.http import HttpRequest, HttpResponse

from ..models import DummyPerson, Person
from .core_helpers import get_site_preferences, has_person


class EnsurePersonMiddleware:
    """Middleware that ensures that the logged-in user is linked to a person.

    It is needed to inject a dummy person to a superuser that would otherwise
    not have an associated person, in order they can get their account set up
    without external help.

    In addition, if configured in preferences, it auto-creates or links persons
    to regular users if they match.
    """

    def __init__(self, get_response: Callable):
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if not has_person(request) and not request.user.is_anonymous:
            prefs = get_site_preferences()
            if (
                prefs.get("account__auto_link_person", False)
                and request.user.first_name
                and request.user.last_name
            ):
                if prefs.get("account__auto_create_person"):
                    person, created = Person.objects.get_or_create(
                        email=request.user.email,
                        defaults={
                            "first_name": request.user.first_name,
                            "last_name": request.user.last_name,
                        },
                    )
                    person.user = request.user
                else:
                    person = Person.objects.filter(email=request.user.email).first()
                    if person:
                        person.user = request.user
                        person.save()

        if request.user.is_superuser and not has_person(request):
            # Super-users get a dummy person linked
            dummy_person = DummyPerson(
                first_name=request.user.first_name, last_name=request.user.last_name
            )
            request.user.person = dummy_person

        response = self.get_response(request)
        return response



import base64

from django.contrib.auth import get_user_model

import pytest

from aleksis.core.models import Group, OAuthApplication, Person

pytestmark = pytest.mark.django_db
from django.http import HttpResponse
from django.test import override_settings
from django.urls import path, reverse
from django.views.generic import View


def test_all_settigns_registered():
    """Tests for regressions of preferences not being registered.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/592
    """

    from dynamic_preferences.types import BasePreferenceType

    from aleksis.core import preferences
    from aleksis.core.preferences import person_preferences_registry, site_preferences_registry

    for obj in preferences.__dict__.values():
        if not isinstance(obj, BasePreferenceType):
            continue

        in_site_reg = site_preferences_registry.get(obj.section.name, {}).get(obj.name, None) is obj
        in_person_reg = (
            person_preferences_registry.get(obj.section.name, {}).get(obj.name, None) is obj
        )

        assert in_site_reg != in_person_reg


def test_custom_managers_return_correct_qs():
    """Tests that custom managers' get_queryset methods return the expected qs.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/594
    """

    from aleksis.core import managers

    def _check_get_queryset(Manager, QuerySet):
        assert isinstance(Manager.from_queryset(QuerySet)().get_queryset(), QuerySet)

    _check_get_queryset(managers.GroupManager, managers.GroupQuerySet)


def test_reassign_user_to_person():
    """Tests that on re-assigning a user, groups are correctly synced.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/628
    """

    User = get_user_model()

    group1 = Group.objects.create(name="Group 1")
    group2 = Group.objects.create(name="Group 2")

    user1 = User.objects.create(username="user1")
    user2 = User.objects.create(username="user2")

    person1 = Person.objects.create(first_name="Person", last_name="1", user=user1)
    person2 = Person.objects.create(first_name="Person", last_name="2", user=user2)

    person1.member_of.set([group1])
    person2.member_of.set([group2])

    assert user1.groups.count() == 1
    assert user2.groups.count() == 1
    assert user1.groups.first().name == "Group 1"
    assert user2.groups.first().name == "Group 2"

    person1.user = None
    person1.save()
    assert user1.groups.count() == 0

    person2.user = user1
    person2.save()
    person1.user = user2
    person1.save()

    assert user1.groups.count() == 1
    assert user2.groups.count() == 1
    assert user1.groups.first().name == "Group 2"
    assert user2.groups.first().name == "Group 1"


@override_settings(ROOT_URLCONF="aleksis.core.tests.regression.view_oauth")
def test_no_access_oauth2_client_credentials_without_allowed_scopes(client):
    """Tests that ClientProtectedResourceMixin doesn't allow access if no allowed scopes are set.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/688
    """

    wrong_application = OAuthApplication(
        name="Test Application",
        allowed_scopes=[],
        authorization_grant_type=OAuthApplication.GRANT_CLIENT_CREDENTIALS,
        client_type=OAuthApplication.CLIENT_CONFIDENTIAL,
        redirect_uris=["http://localhost:8000/"],
    )
    wrong_application_secret = wrong_application.client_secret
    wrong_application.save()
    wrong_application_2 = OAuthApplication(
        name="Test Application",
        allowed_scopes=["read"],
        authorization_grant_type=OAuthApplication.GRANT_CLIENT_CREDENTIALS,
        client_type=OAuthApplication.CLIENT_CONFIDENTIAL,
        redirect_uris=["http://localhost:8000/"],
    )
    wrong_application_2_secret = wrong_application_2.client_secret
    wrong_application_2.save()
    correct_application = OAuthApplication(
        name="Test Application",
        allowed_scopes=["write"],
        authorization_grant_type=OAuthApplication.GRANT_CLIENT_CREDENTIALS,
        client_type=OAuthApplication.CLIENT_CONFIDENTIAL,
        redirect_uris=["http://localhost:8000/"],
    )
    correct_application_secret = correct_application.client_secret
    correct_application.save()

    url = reverse("client_protected_resource_mixin_test")
    auth_header = (
        "Basic "
        + base64.b64encode(
            f"{wrong_application.client_id}:{wrong_application_secret}".encode()
        ).decode()
    )
    r = client.get(url, HTTP_AUTHORIZATION=auth_header)
    assert r.status_code == 403

    auth_header = (
        "Basic "
        + base64.b64encode(
            f"{wrong_application_2.client_id}:{wrong_application_2_secret}".encode()
        ).decode()
    )
    r = client.get(url, HTTP_AUTHORIZATION=auth_header)
    assert r.status_code == 403

    auth_header = (
        "Basic "
        + base64.b64encode(
            f"{correct_application.client_id}:{correct_application_secret}".encode()
        ).decode()
    )
    r = client.get(url, HTTP_AUTHORIZATION=auth_header)
    assert r.status_code == 200


def test_change_password_not_logged_in(client):
    """Tests that CustomPasswordChangeView redirects to login when accessed unauthenticated.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/703
    """
    response = client.get(reverse("account_change_password"), follow=True)

    assert response.status_code == 200
    assert "Please login to see this page." in response.content.decode("utf-8")



# Generated by Django 3.2 on 2021-04-10 18:58

import aleksis.core.models
import django.contrib.sites.managers
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
        ('core', '0012_valid_from_announcement'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='globalpermissions',
            options={'default_permissions': (), 'managed': False, 'permissions': (
            ('view_system_status', 'Can view system status'), ('link_persons_accounts', 'Can link persons to accounts'),
            ('manage_data', 'Can manage data'), ('impersonate', 'Can impersonate'), ('search', 'Can use search'),
            ('change_site_preferences', 'Can change site preferences'),
            ('change_person_preferences', 'Can change person preferences'),
            ('change_group_preferences', 'Can change group preferences'), ('test_pdf', 'Can test PDF generation'))},
        ),
        migrations.CreateModel(
            name='PDFFile',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('expires_at', models.DateTimeField(default=aleksis.core.models.PDFFile._get_default_expiration, verbose_name='File expires at')),
                ('html', models.TextField(verbose_name='Rendered HTML')),
                ('file', models.FileField(blank=True, null=True, upload_to='pdfs/', verbose_name='Generated PDF file')),
                ('person', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='pdf_files', to='core.person', verbose_name='Owner')),
                ('site', models.ForeignKey(default=1, editable=False, on_delete=django.db.models.deletion.CASCADE, to='sites.site')),
            ],
            options={
                'verbose_name': 'PDF file',
                'verbose_name_plural': 'PDF files',
            },
            managers=[
                ('objects', django.contrib.sites.managers.CurrentSiteManager()),
            ],
        ),
    ]



import json
import os
import shutil

from django.conf import settings

from django_yarnpkg.management.base import BaseYarnCommand
from django_yarnpkg.yarn import yarn_adapter

from ...util.frontend_helpers import get_apps_with_assets


class Command(BaseYarnCommand):
    help = "Create webpack bundles for AlekSIS"  # noqa

    def handle(self, *args, **options):
        super(Command, self).handle(*args, **options)

        # Write webpack entrypoints for all apps
        assets = {
            app: {"dependOn": "core", "import": os.path.join(path, "index")}
            for app, path in get_apps_with_assets().items()
        }
        assets["core"] = os.path.join(settings.BASE_DIR, "aleksis", "core", "assets", "index")
        with open(os.path.join(settings.NODE_MODULES_ROOT, "webpack-entrypoints.json"), "w") as out:
            json.dump(assets, out)

        # Install Node dependencies
        yarn_adapter.install(settings.YARN_INSTALLED_APPS)

        # Run webpack
        config_path = os.path.join(settings.BASE_DIR, "aleksis", "core", "webpack.config.js")
        shutil.copy(config_path, settings.NODE_MODULES_ROOT)
        mode = "development" if settings.DEBUG else "production"
        yarn_adapter.call_yarn(["run", "webpack", f"--mode={mode}"])



# Generated by Django 3.2.8 on 2021-11-02 20:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0026_oauthapplication_allowed_scopes'),
    ]

    operations = [
        migrations.AddField(
            model_name='person',
            name='place_of_birth',
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name='Place of birth'),
        ),
    ]



from django.contrib.auth.models import Group as DjangoGroup
from django.contrib.auth.models import User

import pytest

from aleksis.core.models import Group, Person

pytestmark = pytest.mark.django_db


def test_create():
    Group.objects.create(name="Foo")

    assert DjangoGroup.objects.filter(name="Foo").exists()


def test_assign_members():
    g = Group.objects.create(name="Foo")
    dj_g = DjangoGroup.objects.get(name="Foo")

    u = User.objects.create(username="janedoe")
    p = Person.objects.create(first_name="Jane", last_name="Doe", user=u)

    g.members.add(p)

    assert u in dj_g.user_set.all()


def test_assign_owners():
    g = Group.objects.create(name="Foo")
    dj_g = DjangoGroup.objects.get(name="Foo")

    u = User.objects.create(username="janedoe")
    p = Person.objects.create(first_name="Jane", last_name="Doe", user=u)

    g.owners.add(p)

    assert u in dj_g.user_set.all()


def test_assign_member_of():
    g = Group.objects.create(name="Foo")
    dj_g = DjangoGroup.objects.get(name="Foo")

    u = User.objects.create(username="janedoe")
    p = Person.objects.create(first_name="Jane", last_name="Doe", user=u)

    p.member_of.add(g)

    assert u in dj_g.user_set.all()


def test_assign_owner_of():
    g = Group.objects.create(name="Foo")
    dj_g = DjangoGroup.objects.get(name="Foo")

    u = User.objects.create(username="janedoe")
    p = Person.objects.create(first_name="Jane", last_name="Doe", user=u)

    p.owner_of.add(g)

    assert u in dj_g.user_set.all()



# flake8: noqa: DJ01
import base64
import hmac
import uuid
from datetime import date, datetime, timedelta
from typing import Any, Iterable, List, Optional, Sequence, Union
from urllib.parse import urlparse

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group as DjangoGroup
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import ArrayField
from django.contrib.sites.models import Site
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator
from django.db import models, transaction
from django.db.models import Q, QuerySet
from django.db.models.signals import m2m_changed
from django.dispatch import receiver
from django.forms.widgets import Media
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import classproperty
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

import customidenticon
import jsonstore
from cachalot.api import cachalot_disabled
from cache_memoize import cache_memoize
from celery.result import AsyncResult
from ckeditor.fields import RichTextField
from django_celery_results.models import TaskResult
from django_cte import CTEQuerySet, With
from dynamic_preferences.models import PerInstancePreferenceModel
from invitations import signals
from invitations.adapters import get_invitations_adapter
from invitations.base_invitation import AbstractBaseInvitation
from invitations.models import Invitation
from model_utils import FieldTracker
from model_utils.models import TimeStampedModel
from oauth2_provider.models import (
    AbstractAccessToken,
    AbstractApplication,
    AbstractGrant,
    AbstractIDToken,
    AbstractRefreshToken,
)
from phonenumber_field.modelfields import PhoneNumberField
from polymorphic.models import PolymorphicModel

from aleksis.core.data_checks import (
    BrokenDashboardWidgetDataCheck,
    DataCheck,
    DataCheckRegistry,
    field_validation_data_check_factory,
)

from .feeds import PersonalICalFeedBase
from .managers import (
    CurrentSiteManagerWithoutMigrations,
    GroupManager,
    GroupQuerySet,
    InstalledWidgetsDashboardWidgetOrderManager,
    SchoolTermQuerySet,
    UninstallRenitentPolymorphicManager,
)
from .mixins import (
    ExtensibleModel,
    GlobalPermissionModel,
    PureDjangoModel,
    SchoolTermRelatedExtensibleModel,
)
from .tasks import send_notification
from .util.core_helpers import generate_random_code, get_site_preferences, now_tomorrow
from .util.email import send_email
from .util.model_helpers import ICONS

FIELD_CHOICES = (
    ("BooleanField", _("Boolean (Yes/No)")),
    ("CharField", _("Text (one line)")),
    ("DateField", _("Date")),
    ("DateTimeField", _("Date and time")),
    ("DecimalField", _("Decimal number")),
    ("EmailField", _("E-mail address")),
    ("IntegerField", _("Integer")),
    ("GenericIPAddressField", _("IP address")),
    ("NullBooleanField", _("Boolean or empty (Yes/No/Neither)")),
    ("TextField", _("Text (multi-line)")),
    ("TimeField", _("Time")),
    ("URLField", _("URL / Link")),
)


class SchoolTerm(ExtensibleModel):
    """School term model.

    This is used to manage start and end times of a school term and link data to it.
    """

    objects = CurrentSiteManagerWithoutMigrations.from_queryset(SchoolTermQuerySet)()

    name = models.CharField(verbose_name=_("Name"), max_length=255)

    date_start = models.DateField(verbose_name=_("Start date"))
    date_end = models.DateField(verbose_name=_("End date"))

    @classmethod
    @cache_memoize(3600)
    def get_current(cls, day: Optional[date] = None):
        if not day:
            day = timezone.now().date()
        try:
            return cls.objects.on_day(day).first()
        except SchoolTerm.DoesNotExist:
            return None

    @classproperty
    def current(cls):
        return cls.get_current()

    def clean(self):
        """Ensure there is only one school term at each point of time."""
        if self.date_end < self.date_start:
            raise ValidationError(_("The start date must be earlier than the end date."))

        qs = SchoolTerm.objects.within_dates(self.date_start, self.date_end)
        if self.pk:
            qs = qs.exclude(pk=self.pk)
        if qs.exists():
            raise ValidationError(
                _("There is already a school term for this time or a part of this time.")
            )

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _("School term")
        verbose_name_plural = _("School terms")
        constraints = [
            models.UniqueConstraint(
                fields=["site_id", "name"], name="unique_school_term_name_per_site"
            ),
            models.UniqueConstraint(
                fields=["site_id", "date_start", "date_end"],
                name="unique_school_term_dates_per_site",
            ),
        ]


class Person(ExtensibleModel):
    """Person model.

    A model describing any person related to a school, including, but not
    limited to, students, teachers and guardians (parents).
    """

    class Meta:
        ordering = ["last_name", "first_name"]
        verbose_name = _("Person")
        verbose_name_plural = _("Persons")
        permissions = (
            ("view_address", _("Can view address")),
            ("view_contact_details", _("Can view contact details")),
            ("view_photo", _("Can view photo")),
            ("view_avatar", _("Can view avatar image")),
            ("view_person_groups", _("Can view persons groups")),
            ("view_personal_details", _("Can view personal details")),
        )
        constraints = [
            models.UniqueConstraint(
                fields=["site_id", "short_name"], name="unique_short_name_per_site"
            ),
        ]

    icon_ = "person"

    SEX_CHOICES = [("f", _("female")), ("m", _("male")), ("x", _("other"))]

    user = models.OneToOneField(
        get_user_model(),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="person",
        verbose_name=_("Linked user"),
    )

    first_name = models.CharField(verbose_name=_("First name"), max_length=255)
    last_name = models.CharField(verbose_name=_("Last name"), max_length=255)
    additional_name = models.CharField(
        verbose_name=_("Additional name(s)"), max_length=255, blank=True
    )

    short_name = models.CharField(
        verbose_name=_("Short name"), max_length=255, blank=True, null=True  # noqa
    )

    street = models.CharField(verbose_name=_("Street"), max_length=255, blank=True)
    housenumber = models.CharField(verbose_name=_("Street number"), max_length=255, blank=True)
    postal_code = models.CharField(verbose_name=_("Postal code"), max_length=255, blank=True)
    place = models.CharField(verbose_name=_("Place"), max_length=255, blank=True)

    phone_number = PhoneNumberField(verbose_name=_("Home phone"), blank=True)
    mobile_number = PhoneNumberField(verbose_name=_("Mobile phone"), blank=True)

    email = models.EmailField(verbose_name=_("E-mail address"), blank=True)

    date_of_birth = models.DateField(verbose_name=_("Date of birth"), blank=True, null=True)
    place_of_birth = models.CharField(verbose_name=_("Place of birth"), max_length=255, blank=True)
    sex = models.CharField(verbose_name=_("Sex"), max_length=1, choices=SEX_CHOICES, blank=True)

    photo = models.ImageField(
        verbose_name=_("Photo"),
        blank=True,
        null=True,
        help_text=_(
            "This is an official photo, used for official documents and for internal use cases."
        ),
    )

    avatar = models.ImageField(
        verbose_name=_("Display picture / Avatar"),
        blank=True,
        null=True,
        help_text=_("This is a picture or an avatar for public display."),
    )

    guardians = models.ManyToManyField(
        "self",
        verbose_name=_("Guardians / Parents"),
        symmetrical=False,
        related_name="children",
        blank=True,
    )

    primary_group = models.ForeignKey(
        "Group", models.SET_NULL, null=True, blank=True, verbose_name=_("Primary group")
    )

    description = models.TextField(verbose_name=_("Description"), blank=True)

    def get_absolute_url(self) -> str:
        return reverse("person_by_id", args=[self.id])

    @property
    def primary_group_short_name(self) -> Optional[str]:
        """Return the short_name field of the primary group related object."""
        if self.primary_group:
            return self.primary_group.short_name

    @primary_group_short_name.setter
    def primary_group_short_name(self, value: str) -> None:
        """
        Set the primary group related object by a short name.

        It uses the first existing group
        with this short name it can find, creating one
        if it can't find one.
        """
        group, created = Group.objects.get_or_create(short_name=value, defaults={"name": value})
        self.primary_group = group

    @property
    def full_name(self) -> str:
        """Full name of person in last name, first name order."""
        return f"{self.last_name}, {self.first_name}"

    @property
    def addressing_name(self) -> str:
        """Full name of person in format configured for addressing."""
        if self.preferences["notification__addressing_name_format"] == "last_first":
            return f"{self.last_name}, {self.first_name}"
        elif self.preferences["notification__addressing_name_format"] == "first_last":
            return f"{self.first_name} {self.last_name}"

    @property
    def mail_sender(self) -> str:
        """E-mail sender in "Name <email>" format."""
        return f'"{self.addressing_name}" <{self.email}>'

    @property
    def mail_sender_via(self) -> str:
        """E-mail sender for via addresses, in "Name via Site <email>" format."""
        site_mail = get_site_preferences()["mail__address"]
        site_name = get_site_preferences()["general__title"]

        return f'"{self.addressing_name} via {site_name}" <{site_mail}>'

    @property
    def age(self):
        """Age of the person at current time."""
        return self.age_at(timezone.now().date())

    def age_at(self, today):
        if self.date_of_birth:
            years = today.year - self.date_of_birth.year
            if self.date_of_birth.month > today.month or (
                self.date_of_birth.month == today.month and self.date_of_birth.day > today.day
            ):
                years -= 1
            return years

    @property
    def dashboard_widgets(self):
        return [
            w.widget
            for w in DashboardWidgetOrder.objects.filter(person=self, widget__active=True).order_by(
                "order"
            )
        ]

    @property
    def unread_notifications(self) -> QuerySet:
        """Get all unread notifications for this person."""
        return self.notifications.filter(read=False, send_at__lte=timezone.now())

    @property
    def unread_notifications_count(self) -> int:
        """Return the count of unread notifications for this person."""
        return self.unread_notifications.count()

    @property
    def initials(self):
        initials = ""
        if self.first_name:
            initials += self.first_name[0]
        if self.last_name:
            initials += self.last_name[0]
        return initials.upper() or "?"

    user_info_tracker = FieldTracker(fields=("first_name", "last_name", "email", "user_id"))

    @property
    def member_of_recursive(self) -> QuerySet:
        """Get all groups this person is a member of, recursively."""
        q = self.member_of
        for group in q.all():
            q = q.union(group.parent_groups_recursive)
        return q

    @property
    def owner_of_recursive(self) -> QuerySet:
        """Get all groups this person is a member of, recursively."""
        q = self.owner_of
        for group in q.all():
            q = q.union(group.child_groups_recursive)
        return q

    @property
    @cache_memoize(60 * 60)
    def identicon_url(self):
        identicon = customidenticon.create(self.full_name, border=35)
        base64_data = base64.b64encode(identicon).decode("ascii")
        return f"data:image/png;base64,{base64_data}"

    @property
    def avatar_url(self):
        if self.avatar:
            return self.avatar.url
        else:
            return self.identicon_url

    def save(self, *args, **kwargs):
        # Determine all fields that were changed since last load
        changed = self.user_info_tracker.changed()

        super().save(*args, **kwargs)

        if self.pk is None or bool(changed):
            if "user_id" in changed:
                # Clear groups of previous Django user
                previous_user = changed["user_id"]
                if previous_user is not None:
                    get_user_model().objects.get(pk=previous_user).groups.clear()

            if self.user:
                if "first_name" in changed or "last_name" in changed or "email" in changed:
                    # Synchronise user fields to linked User object to keep it up to date
                    self.user.first_name = self.first_name
                    self.user.last_name = self.last_name
                    self.user.email = self.email
                    self.user.save()

                if "user_id" in changed:
                    # Synchronise groups to Django groups
                    for group in self.member_of.union(self.owner_of.all()).all():
                        group.save(force=True)

        # Select a primary group if none is set
        self.auto_select_primary_group()

    def __str__(self) -> str:
        return self.full_name

    @classmethod
    def maintain_default_data(cls):
        # Ensure we have an admin user
        user = get_user_model()
        if not user.objects.filter(is_superuser=True).exists():
            admin = user.objects.create_superuser(**settings.AUTH_INITIAL_SUPERUSER)
            admin.save()

    def auto_select_primary_group(
        self, pattern: Optional[str] = None, field: Optional[str] = None, force: bool = False
    ) -> None:
        """Auto-select the primary group among the groups the person is member of.

        Uses either the pattern passed as argument, or the pattern configured system-wide.

        Does not do anything if either no pattern is defined or the user already has
        a primary group, unless force is True.
        """
        pattern = pattern or get_site_preferences()["account__primary_group_pattern"]
        field = field or get_site_preferences()["account__primary_group_field"]

        if pattern:
            if force or not self.primary_group:
                self.primary_group = self.member_of.filter(**{f"{field}__regex": pattern}).first()

    def notify_about_changed_data(
        self, changed_fields: Iterable[str], recipients: Optional[List[str]] = None
    ):
        """Notify (configured) recipients about changed data of this person."""
        context = {"person": self, "changed_fields": changed_fields}
        recipients = recipients or [
            get_site_preferences()["account__person_change_notification_contact"]
        ]
        send_email(
            template_name="person_changed",
            from_email=self.mail_sender_via,
            headers={
                "Reply-To": self.mail_sender,
                "Sender": self.mail_sender,
            },
            recipient_list=recipients,
            context=context,
        )


class DummyPerson(Person):
    """A dummy person that is not stored into the database.

    Used to temporarily inject a Person object into a User.
    """

    class Meta:
        managed = False
        proxy = True

    is_dummy = True

    def save(self, *args, **kwargs):
        # Do nothing, not even call Model's save(), so this is never persisted
        pass


class AdditionalField(ExtensibleModel):
    """An additional field that can be linked to a group."""

    title = models.CharField(verbose_name=_("Title of field"), max_length=255)
    field_type = models.CharField(
        verbose_name=_("Type of field"), choices=FIELD_CHOICES, max_length=50
    )
    required = models.BooleanField(verbose_name=_("Required"), default=False)
    help_text = models.TextField(verbose_name=_("Help text / description"), blank=True)

    def __str__(self) -> str:
        return self.title

    class Meta:
        verbose_name = _("Addtitional field for groups")
        verbose_name_plural = _("Addtitional fields for groups")
        constraints = [
            models.UniqueConstraint(fields=["site_id", "title"], name="unique_title_per_site"),
        ]


class Group(SchoolTermRelatedExtensibleModel):
    """Group model.

    Any kind of group of persons in a school, including, but not limited
    classes, clubs, and the like.
    """

    objects = GroupManager.from_queryset(GroupQuerySet)()

    class Meta:
        ordering = ["short_name", "name"]
        verbose_name = _("Group")
        verbose_name_plural = _("Groups")
        permissions = (
            ("assign_child_groups_to_groups", _("Can assign child groups to groups")),
            ("view_group_stats", _("Can view statistics about group.")),
        )
        constraints = [
            # Heads up: Uniqueness per school term already implies uniqueness per site
            models.UniqueConstraint(fields=["school_term", "name"], name="unique_school_term_name"),
            models.UniqueConstraint(
                fields=["school_term", "short_name"], name="unique_school_term_short_name"
            ),
        ]

    icon_ = "group"

    name = models.CharField(verbose_name=_("Long name"), max_length=255)
    short_name = models.CharField(
        verbose_name=_("Short name"), max_length=255, blank=True, null=True  # noqa
    )

    members = models.ManyToManyField(
        "Person",
        related_name="member_of",
        blank=True,
        through="PersonGroupThrough",
        verbose_name=_("Members"),
    )
    owners = models.ManyToManyField(
        "Person", related_name="owner_of", blank=True, verbose_name=_("Owners")
    )

    parent_groups = models.ManyToManyField(
        "self",
        related_name="child_groups",
        symmetrical=False,
        verbose_name=_("Parent groups"),
        blank=True,
    )

    group_type = models.ForeignKey(
        "GroupType",
        on_delete=models.SET_NULL,
        related_name="type",
        verbose_name=_("Type of group"),
        null=True,
        blank=True,
    )
    additional_fields = models.ManyToManyField(
        AdditionalField, verbose_name=_("Additional fields"), blank=True
    )

    photo = models.ImageField(
        verbose_name=_("Photo"),
        blank=True,
        null=True,
        help_text=_(
            "This is an official photo, used for official documents and for internal use cases."
        ),
    )
    avatar = models.ImageField(
        verbose_name=_("Display picture / Avatar"),
        blank=True,
        null=True,
        help_text=_("This is a picture or an avatar for public display."),
    )

    def get_absolute_url(self) -> str:
        return reverse("group_by_id", args=[self.id])

    @property
    def announcement_recipients(self):
        """Flat list of all members and owners to fulfill announcement API contract."""
        return list(self.members.all()) + list(self.owners.all())

    @property
    def get_group_stats(self) -> dict:
        """Get stats about a given group"""
        stats = {}

        stats["members"] = len(self.members.all())

        ages = [person.age for person in self.members.filter(date_of_birth__isnull=False)]

        if ages:
            stats["age_avg"] = sum(ages) / len(ages)
            stats["age_range_min"] = min(ages)
            stats["age_range_max"] = max(ages)

        return stats

    @property
    def parent_groups_recursive(self) -> CTEQuerySet:
        """Get all parent groups recursively."""

        def _make_cte(cte):
            Through = self.parent_groups.through
            return (
                Through.objects.values("to_group_id")
                .filter(from_group=self)
                .union(cte.join(Through, from_group=cte.col.to_group_id), all=True)
            )

        cte = With.recursive(_make_cte)
        return cte.join(Group, id=cte.col.to_group_id).with_cte(cte)

    @property
    def child_groups_recursive(self) -> CTEQuerySet:
        """Get all child groups recursively."""

        def _make_cte(cte):
            Through = self.child_groups.through
            return (
                Through.objects.values("from_group_id")
                .filter(to_group=self)
                .union(cte.join(Through, to_group=cte.col.from_group_id), all=True)
            )

        cte = With.recursive(_make_cte)
        return cte.join(Group, id=cte.col.from_group_id).with_cte(cte)

    @property
    def members_recursive(self) -> QuerySet:
        """Get all members of this group and its child groups."""
        return Person.objects.filter(
            Q(member_of=self) | Q(member_of__in=self.child_groups_recursive)
        )

    @property
    def owners_recursive(self) -> QuerySet:
        """Get all ownerss of this group and its parent groups."""
        return Person.objects.filter(
            Q(owner_of=self) | Q(owner_of__in=self.parent_groups_recursive)
        )

    def __str__(self) -> str:
        if self.school_term:
            return f"{self.name} ({self.short_name}) ({self.school_term})"
        else:
            return f"{self.name} ({self.short_name})"

    group_info_tracker = FieldTracker(fields=("name", "short_name"))

    def save(self, force: bool = False, *args, **kwargs):
        # Determine state of object in relation to database
        dirty = self.pk is None or bool(self.group_info_tracker.changed())

        super().save(*args, **kwargs)

        if force or dirty:
            # Synchronise group to Django group with same name
            dj_group, _ = DjangoGroup.objects.get_or_create(name=self.name)
            dj_group.user_set.set(
                list(
                    self.members.filter(user__isnull=False)
                    .values_list("user", flat=True)
                    .union(self.owners.filter(user__isnull=False).values_list("user", flat=True))
                )
            )
            dj_group.save()

    @property
    def django_group(self):
        """Get Django group for this group."""
        dj_group, _ = DjangoGroup.objects.get_or_create(name=self.name)
        return dj_group


class PersonGroupThrough(ExtensibleModel):
    """Through table for many-to-many relationship of group members.

    It does not have any fields on its own; these are generated upon instantiation
    by inspecting the additional fields selected for the linked group.
    """

    group = models.ForeignKey(Group, on_delete=models.CASCADE)
    person = models.ForeignKey(Person, on_delete=models.CASCADE)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for field in self.group.additional_fields.all():
            field_class = getattr(jsonstore, field.field_type)
            field_name = slugify(field.title).replace("-", "_")
            field_instance = field_class(verbose_name=field.title)
            setattr(self, field_name, field_instance)


@receiver(models.signals.m2m_changed, sender=PersonGroupThrough)
@receiver(models.signals.m2m_changed, sender=Group.owners.through)
def save_group_on_m2m_changed(
    sender: Union[PersonGroupThrough, Group.owners.through],
    instance: models.Model,
    action: str,
    reverse: bool,
    model: models.Model,
    pk_set: Optional[set],
    **kwargs,
) -> None:
    """Ensure user and group data is synced to Django's models.

    AlekSIS maintains personal information and group meta-data / membership
    in its Person and Group models. As third-party libraries have no knowledge
    about this, we need to keep django.contrib.auth in sync.

    This signal handler triggers a save of group objects whenever a membership
    changes. The save() code will decide whether to update the Django objects
    or not.
    """
    if action not in ("post_add", "post_remove", "post_clear"):
        # Only trigger once, after the change was applied to the database
        return

    if reverse:
        # Relationship was changed on the Person side, saving all groups
        # that have been touched there
        for group in model.objects.filter(pk__in=pk_set):
            group.save(force=True)
    else:
        # Relationship was changed on the Group side
        instance.save(force=True)


class Activity(ExtensibleModel, TimeStampedModel):
    """Activity of a user to trace some actions done in AlekSIS in displayable form."""

    user = models.ForeignKey(
        "Person", on_delete=models.CASCADE, related_name="activities", verbose_name=_("User")
    )

    title = models.CharField(max_length=150, verbose_name=_("Title"))
    description = models.TextField(max_length=500, verbose_name=_("Description"))

    app = models.CharField(max_length=100, verbose_name=_("Application"))

    def __str__(self):
        return self.title

    class Meta:
        verbose_name = _("Activity")
        verbose_name_plural = _("Activities")


class Notification(ExtensibleModel, TimeStampedModel):
    """Notification to submit to a user."""

    sender = models.CharField(max_length=100, verbose_name=_("Sender"))
    recipient = models.ForeignKey(
        "Person",
        on_delete=models.CASCADE,
        related_name="notifications",
        verbose_name=_("Recipient"),
    )

    title = models.CharField(max_length=150, verbose_name=_("Title"))
    description = models.TextField(max_length=500, verbose_name=_("Description"))
    link = models.URLField(blank=True, verbose_name=_("Link"))

    send_at = models.DateTimeField(default=timezone.now, verbose_name=_("Send notification at"))

    read = models.BooleanField(default=False, verbose_name=_("Read"))
    sent = models.BooleanField(default=False, verbose_name=_("Sent"))

    def __str__(self):
        return str(self.title)

    def save(self, **kwargs):
        super().save(**kwargs)
        if not self.sent and self.send_at <= timezone.now():
            self.send()
            super().save(**kwargs)

    def send(self, resend: bool = False) -> Optional[AsyncResult]:
        """Send the notification to the recipient."""
        if not self.sent or resend:
            return send_notification.delay(self.pk, resend=True)

    class Meta:
        verbose_name = _("Notification")
        verbose_name_plural = _("Notifications")


class AnnouncementQuerySet(models.QuerySet):
    """Queryset for announcements providing time-based utility functions."""

    def relevant_for(self, obj: Union[models.Model, models.QuerySet]) -> models.QuerySet:
        """Get all relevant announcements.

        Get a QuerySet with all announcements relevant for a certain Model (e.g. a Group)
        or a set of models in a QuerySet.
        """
        if isinstance(obj, models.QuerySet):
            ct = ContentType.objects.get_for_model(obj.model)
            pks = list(obj.values_list("pk", flat=True))
        else:
            ct = ContentType.objects.get_for_model(obj)
            pks = [obj.pk]

        return self.filter(recipients__content_type=ct, recipients__recipient_id__in=pks)

    def at_time(self, when: Optional[datetime] = None) -> models.QuerySet:
        """Get all announcements at a certain time."""
        when = when or timezone.now()

        # Get announcements by time
        announcements = self.filter(valid_from__lte=when, valid_until__gte=when)

        return announcements

    def on_date(self, when: Optional[date] = None) -> models.QuerySet:
        """Get all announcements at a certain date."""
        when = when or timezone.now().date()

        # Get announcements by time
        announcements = self.filter(valid_from__date__lte=when, valid_until__date__gte=when)

        return announcements

    def within_days(self, start: date, stop: date) -> models.QuerySet:
        """Get all announcements valid for a set of days."""
        # Get announcements
        announcements = self.filter(valid_from__date__lte=stop, valid_until__date__gte=start)

        return announcements

    def for_person(self, person: Person) -> List:
        """Get all announcements for one person."""
        # Filter by person
        announcements_for_person = []
        for announcement in self:
            if person in announcement.recipient_persons:
                announcements_for_person.append(announcement)

        return announcements_for_person


class Announcement(ExtensibleModel):
    """Announcement model.

    Persistent announcement to display to groups or persons in various places during a
    specific time range.
    """

    objects = models.Manager.from_queryset(AnnouncementQuerySet)()

    title = models.CharField(max_length=150, verbose_name=_("Title"))
    description = models.TextField(max_length=500, verbose_name=_("Description"), blank=True)
    link = models.URLField(blank=True, verbose_name=_("Link to detailed view"))

    valid_from = models.DateTimeField(
        verbose_name=_("Date and time from when to show"), default=timezone.now
    )
    valid_until = models.DateTimeField(
        verbose_name=_("Date and time until when to show"),
        default=now_tomorrow,
    )

    @property
    def recipient_persons(self) -> Sequence[Person]:
        """Return a list of Persons this announcement is relevant for."""
        persons = []
        for recipient in self.recipients.all():
            persons += recipient.persons
        return persons

    def get_recipients_for_model(self, obj: Union[models.Model]) -> Sequence[models.Model]:
        """Get all recipients.

        Get all recipients for this announcement
        with a special content type (provided through model)
        """
        ct = ContentType.objects.get_for_model(obj)
        return [r.recipient for r in self.recipients.filter(content_type=ct)]

    def __str__(self):
        return self.title

    class Meta:
        verbose_name = _("Announcement")
        verbose_name_plural = _("Announcements")


class AnnouncementRecipient(ExtensibleModel):
    """Announcement recipient model.

    Generalisation of a recipient for an announcement, used to wrap arbitrary
    objects that can receive announcements.

    Contract: Objects to serve as recipient have a property announcement_recipients
    returning a flat list of Person objects.
    """

    announcement = models.ForeignKey(
        Announcement, on_delete=models.CASCADE, related_name="recipients"
    )

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    recipient_id = models.PositiveIntegerField()
    recipient = GenericForeignKey("content_type", "recipient_id")

    @property
    def persons(self) -> Sequence[Person]:
        """Return a list of Persons selected by this recipient object.

        If the recipient is a Person, return that object. If not, it returns the list
        from the announcement_recipients field on the target model.
        """
        if isinstance(self.recipient, Person):
            return [self.recipient]
        else:
            return getattr(self.recipient, "announcement_recipients", [])

    def __str__(self):
        return str(self.recipient)

    class Meta:
        verbose_name = _("Announcement recipient")
        verbose_name_plural = _("Announcement recipients")


class DashboardWidget(PolymorphicModel, PureDjangoModel):
    """Base class for dashboard widgets on the index page."""

    objects = UninstallRenitentPolymorphicManager()

    data_checks = [BrokenDashboardWidgetDataCheck]

    @staticmethod
    def get_media(widgets: Union[QuerySet, Iterable]):
        """Return all media required to render the selected widgets."""
        media = Media()
        for widget in widgets:
            media = media + widget.media
        return media

    template = None
    template_broken = "core/dashboard_widget/dashboardwidget_broken.html"
    media = Media()

    title = models.CharField(max_length=150, verbose_name=_("Widget Title"))
    active = models.BooleanField(verbose_name=_("Activate Widget"))
    broken = models.BooleanField(verbose_name=_("Widget is broken"), default=False)

    size_s = models.PositiveSmallIntegerField(
        verbose_name=_("Size on mobile devices"),
        help_text=_("<= 600 px, 12 columns"),
        validators=[MaxValueValidator(12)],
        default=12,
    )
    size_m = models.PositiveSmallIntegerField(
        verbose_name=_("Size on tablet devices"),
        help_text=_("> 600 px, 12 columns"),
        validators=[MaxValueValidator(12)],
        default=12,
    )
    size_l = models.PositiveSmallIntegerField(
        verbose_name=_("Size on desktop devices"),
        help_text=_("> 992 px, 12 columns"),
        validators=[MaxValueValidator(12)],
        default=6,
    )
    size_xl = models.PositiveSmallIntegerField(
        verbose_name=_("Size on large desktop devices"),
        help_text=_("> 1200 px>, 12 columns"),
        validators=[MaxValueValidator(12)],
        default=4,
    )

    def _get_context_safe(self, request):
        if self.broken:
            return {"title": self.title}
        return self.get_context(request)

    def get_context(self, request):
        """Get the context dictionary to pass to the widget template."""
        raise NotImplementedError("A widget subclass needs to implement the get_context method.")

    def get_template(self):
        """Get template.

        Get the template to render the widget with. Defaults to the template attribute,
        but can be overridden to allow more complex template generation scenarios. If
        the widget is marked as broken, the template_broken attribute will be returned.
        """
        if self.broken:
            return self.template_broken
        if not self.template:
            raise NotImplementedError("A widget subclass needs to define a template.")
        return self.template

    def __str__(self):
        return self.title

    class Meta:
        permissions = (("edit_default_dashboard", _("Can edit default dashboard")),)
        verbose_name = _("Dashboard Widget")
        verbose_name_plural = _("Dashboard Widgets")


class ExternalLinkWidget(DashboardWidget):
    template = "core/dashboard_widget/external_link_widget.html"

    url = models.URLField(verbose_name=_("URL"))
    icon_url = models.URLField(verbose_name=_("Icon URL"))

    def get_context(self, request):
        return {"title": self.title, "url": self.url, "icon_url": self.icon_url}

    class Meta:
        verbose_name = _("External link widget")
        verbose_name_plural = _("External link widgets")


class StaticContentWidget(DashboardWidget):
    template = "core/dashboard_widget/static_content_widget.html"

    content = RichTextField(verbose_name=_("Content"))

    def get_context(self, request):
        return {"title": self.title, "content": self.content}

    class Meta:
        verbose_name = _("Static content widget")
        verbose_name_plural = _("Static content widgets")


class DashboardWidgetOrder(ExtensibleModel):
    widget = models.ForeignKey(
        DashboardWidget, on_delete=models.CASCADE, verbose_name=_("Dashboard widget")
    )
    person = models.ForeignKey(
        Person, on_delete=models.CASCADE, verbose_name=_("Person"), null=True, blank=True
    )
    order = models.PositiveIntegerField(verbose_name=_("Order"))
    default = models.BooleanField(default=False, verbose_name=_("Part of the default dashboard"))

    objects = InstalledWidgetsDashboardWidgetOrderManager()

    @classproperty
    def default_dashboard_widgets(cls):
        """Get default order for dashboard widgets."""
        return [
            w.widget
            for w in cls.objects.filter(person=None, default=True, widget__active=True).order_by(
                "order"
            )
        ]

    class Meta:
        verbose_name = _("Dashboard widget order")
        verbose_name_plural = _("Dashboard widget orders")


class CustomMenu(ExtensibleModel):
    """A custom menu to display in the footer."""

    name = models.CharField(max_length=100, verbose_name=_("Menu ID"))

    def __str__(self):
        return self.name if self.name != "" else self.id

    @classmethod
    @cache_memoize(3600)
    def get_default(cls, name):
        """Get a menu by name or create if it does not exist."""
        menu, _ = cls.objects.prefetch_related("items").get_or_create(name=name)
        return menu

    class Meta:
        verbose_name = _("Custom menu")
        verbose_name_plural = _("Custom menus")
        constraints = [
            models.UniqueConstraint(fields=["site_id", "name"], name="unique_menu_name_per_site"),
        ]


class CustomMenuItem(ExtensibleModel):
    """Single item in a custom menu."""

    data_checks = [field_validation_data_check_factory("core", "CustomMenuItem", "icon")]

    menu = models.ForeignKey(
        CustomMenu, models.CASCADE, verbose_name=_("Menu"), related_name="items"
    )
    name = models.CharField(max_length=150, verbose_name=_("Name"))
    url = models.URLField(verbose_name=_("Link"))
    icon = models.CharField(max_length=50, blank=True, choices=ICONS, verbose_name=_("Icon"))

    def __str__(self):
        return f"[{self.menu}] {self.name}"

    class Meta:
        verbose_name = _("Custom menu item")
        verbose_name_plural = _("Custom menu items")
        constraints = [
            # Heads up: Uniqueness per menu already implies uniqueness per site
            models.UniqueConstraint(fields=["menu", "name"], name="unique_name_per_menu"),
        ]

    def get_absolute_url(self):
        return reverse("admin:core_custommenuitem_change", args=[self.id])


class GroupType(ExtensibleModel):
    """Group type model.

    Descriptive type of a group; used to tag groups and for apps to distinguish
    how to display or handle a certain group.
    """

    name = models.CharField(verbose_name=_("Title of type"), max_length=50)
    description = models.CharField(verbose_name=_("Description"), max_length=500)

    def __str__(self) -> str:
        return self.name

    class Meta:
        verbose_name = _("Group type")
        verbose_name_plural = _("Group types")
        constraints = [
            models.UniqueConstraint(
                fields=["site_id", "name"], name="unique_group_type_name_per_site"
            ),
        ]


class GlobalPermissions(GlobalPermissionModel):
    """Container for global permissions."""

    class Meta(GlobalPermissionModel.Meta):
        permissions = (
            ("view_system_status", _("Can view system status")),
            ("manage_data", _("Can manage data")),
            ("impersonate", _("Can impersonate")),
            ("search", _("Can use search")),
            ("change_site_preferences", _("Can change site preferences")),
            ("change_person_preferences", _("Can change person preferences")),
            ("change_group_preferences", _("Can change group preferences")),
            ("test_pdf", _("Can test PDF generation")),
            ("invite", _("Can invite persons")),
        )


class SitePreferenceModel(PerInstancePreferenceModel, PureDjangoModel):
    """Preference model to hold pereferences valid for a site."""

    instance = models.ForeignKey(Site, on_delete=models.CASCADE)

    class Meta(PerInstancePreferenceModel.Meta):
        app_label = "core"


class PersonPreferenceModel(PerInstancePreferenceModel, PureDjangoModel):
    """Preference model to hold pereferences valid for a person."""

    instance = models.ForeignKey(Person, on_delete=models.CASCADE)

    class Meta(PerInstancePreferenceModel.Meta):
        app_label = "core"


class GroupPreferenceModel(PerInstancePreferenceModel, PureDjangoModel):
    """Preference model to hold pereferences valid for members of a group."""

    instance = models.ForeignKey(Group, on_delete=models.CASCADE)

    class Meta(PerInstancePreferenceModel.Meta):
        app_label = "core"


class DataCheckResult(ExtensibleModel):
    """Save the result of a data check for a specific object."""

    check = models.CharField(
        max_length=255,
        verbose_name=_("Related data check task"),
        choices=DataCheckRegistry.data_checks_choices,
    )

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.CharField(max_length=255)
    related_object = GenericForeignKey("content_type", "object_id")

    solved = models.BooleanField(default=False, verbose_name=_("Issue solved"))
    sent = models.BooleanField(default=False, verbose_name=_("Notification sent"))

    @property
    def related_check(self) -> DataCheck:
        return DataCheckRegistry.data_checks_by_name[self.check]

    def solve(self, solve_option: str = "default"):
        self.related_check.solve(self, solve_option)

    def __str__(self):
        return f"{self.related_object}: {self.related_check.problem_name}"

    class Meta:
        verbose_name = _("Data check result")
        verbose_name_plural = _("Data check results")
        permissions = (
            ("run_data_checks", _("Can run data checks")),
            ("solve_data_problem", _("Can solve data check problems")),
        )


class PersonInvitation(AbstractBaseInvitation, PureDjangoModel):
    """Custom model for invitations to allow to generate invitations codes without email address."""

    email = models.EmailField(verbose_name=_("E-Mail address"), blank=True)
    person = models.ForeignKey(
        Person, on_delete=models.CASCADE, blank=True, related_name="invitation", null=True
    )

    def __str__(self) -> str:
        return f"{self.email} ({self.inviter})"

    @classmethod
    def create(cls, email, inviter=None, **kwargs):
        length = get_site_preferences()["auth__invite_code_length"]
        packet_size = get_site_preferences()["auth__invite_code_packet_size"]
        code = generate_random_code(length, packet_size)

        instance = cls.objects.create(email=email, inviter=inviter, key=code, **kwargs)
        return instance

    key_expired = Invitation.key_expired
    send_invitation = Invitation.send_invitation


class PDFFile(ExtensibleModel):
    """Link to a rendered PDF file."""

    def _get_default_expiration():  # noqa
        return timezone.now() + timedelta(minutes=get_site_preferences()["general__pdf_expiration"])

    person = models.ForeignKey(
        to=Person,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        verbose_name=_("Owner"),
        related_name="pdf_files",
    )
    expires_at = models.DateTimeField(
        verbose_name=_("File expires at"), default=_get_default_expiration
    )
    html_file = models.FileField(
        upload_to="pdfs/", verbose_name=_("Generated HTML file"), blank=True, null=True
    )
    file = models.FileField(
        upload_to="pdfs/", blank=True, null=True, verbose_name=_("Generated PDF file")
    )

    def __str__(self):
        return f"{self.person} ({self.pk})"

    class Meta:
        verbose_name = _("PDF file")
        verbose_name_plural = _("PDF files")


class TaskUserAssignment(ExtensibleModel):
    task_result = models.ForeignKey(
        TaskResult, on_delete=models.CASCADE, verbose_name=_("Task result")
    )
    user = models.ForeignKey(
        get_user_model(), on_delete=models.CASCADE, verbose_name=_("Task user")
    )

    @classmethod
    def create_for_task_id(cls, task_id: str, user: "User") -> "TaskUserAssignment":
        # Use get_or_create to ensure the TaskResult exists
        # django-celery-results will later add the missing information
        with cachalot_disabled():
            result, __ = TaskResult.objects.get_or_create(task_id=task_id)
        return cls.objects.create(task_result=result, user=user)

    class Meta:
        verbose_name = _("Task user assignment")
        verbose_name_plural = _("Task user assignments")


class UserAdditionalAttributes(models.Model, PureDjangoModel):
    """Additional attributes for Django user accounts.

    These attributes are explicitly linked to a User, not to a Person.
    """

    user = models.OneToOneField(
        get_user_model(),
        on_delete=models.CASCADE,
        related_name="additional_attributes",
        verbose_name=_("Linked user"),
    )

    attributes = models.JSONField(verbose_name=_("Additional attributes"), default=dict)

    @classmethod
    def get_user_attribute(
        cls, username: str, attribute: str, default: Optional[Any] = None
    ) -> Any:
        """Get a user attribute for a user by name."""
        try:
            attributes = cls.objects.get(user__username=username)
        except cls.DoesNotExist:
            return default

        return attributes.attributes.get(attribute, default)

    @classmethod
    def set_user_attribute(cls, username: str, attribute: str, value: Any):
        """Set a user attribute for a user by name.

        Raises DoesNotExist if a username for a non-existing Django user is passed.
        """
        user = get_user_model().objects.get(username=username)
        attributes, __ = cls.objects.update_or_create(user=user)

        attributes.attributes[attribute] = value
        attributes.save()


class OAuthApplication(AbstractApplication):
    """Modified OAuth application class that supports Grant Flows configured in preferences."""

    # Override grant types to make field optional
    authorization_grant_type = models.CharField(
        max_length=32, choices=AbstractApplication.GRANT_TYPES, blank=True
    )

    # Optional list of alloewd scopes
    allowed_scopes = ArrayField(
        models.CharField(max_length=255),
        verbose_name=_("Allowed scopes that clients can request"),
        null=True,
        blank=True,
    )

    icon = models.ImageField(
        verbose_name=_("Icon"),
        blank=True,
        null=True,
        help_text=_(
            "This image will be shown as icon in the authorization flow. It should be squared."
        ),
    )

    def allows_grant_type(self, *grant_types: set[str]) -> bool:
        allowed_grants = get_site_preferences()["auth__oauth_allowed_grants"]

        return bool(set(allowed_grants) & set(grant_types))


class OAuthGrant(AbstractGrant):
    """Placeholder for customising the Grant model."""

    pass


class OAuthAccessToken(AbstractAccessToken):
    """Placeholder for customising the AccessToken model."""

    pass


class OAuthIDToken(AbstractIDToken):
    """Placeholder for customising the IDToken model."""

    pass


class OAuthRefreshToken(AbstractRefreshToken):
    """Placeholder for customising the RefreshToken model."""

    pass


class PersonalICalUrl(models.Model):
    """Calendar URL for a person.

    This is used to connect iCalendar subscriptions to a person. A person can have multiple
    URLs. The URL is used to generate the iCalendar feed and have personalized results. It
    is possible to create multiple URLs for the same person and the same iCal feed, e.g. to
    allow a person to share and unshare their calendar with other people.
    """

    person = models.ForeignKey(
        "Person",
        on_delete=models.CASCADE,
        related_name="calendar_urls",
        verbose_name=_("Person"),
    )
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, verbose_name=_("UUID"), unique=True)
    name = models.CharField(max_length=255, verbose_name=_("Name"))
    ical_feed = models.CharField(
        max_length=255,
        verbose_name=_("Selected ICal feed"),
        choices=PersonalICalFeedBase.subclass_choices,
    )

    @property
    def ical_feed_object(self):
        return PersonalICalFeedBase.subclasses_dict.get(self.ical_feed)

    class Meta:
        verbose_name = _("Personal Calendar URL")
        verbose_name_plural = _("Personal Calendar URLs")

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse("ical_feed", kwargs={"slug": self.uuid})



from importlib import metadata

from .celery import app as celery_app  # noqa

try:
    __version__ = metadata.distribution("AlekSIS-Core").version
except Exception:
    __version__ = "unknown"

default_app_config = "aleksis.core.apps.CoreConfig"



# Generated by Django 3.1.7 on 2021-03-03 20:04

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0009_default_dashboard'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExternalLinkWidget',
            fields=[
                ('dashboardwidget_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.dashboardwidget')),
                ('url', models.URLField(verbose_name='URL')),
                ('icon_url', models.URLField(verbose_name='Icon URL')),
            ],
            options={
                'verbose_name': 'External link widget',
                'verbose_name_plural': 'External link widgets',
            },
            bases=('core.dashboardwidget',),
        ),
    ]



import os

from django.conf import settings

from .core_helpers import get_app_module, get_app_packages


def get_apps_with_assets():
    """Get a dictionary of apps that ship frontend assets."""
    assets = {}
    for app in get_app_packages():
        mod = get_app_module(app, "apps")
        path = os.path.join(os.path.dirname(mod.__file__), "assets")
        if os.path.isdir(path):
            package = ".".join(app.split(".")[:-2])
            assets[package] = path
    return assets


def get_language_cookie(code: str) -> str:
    """Build a cookie string to set a new language."""
    cookie_parts = [f"{settings.LANGUAGE_COOKIE_NAME}={code}"]
    args = dict(
        max_age=settings.LANGUAGE_COOKIE_AGE,
        path=settings.LANGUAGE_COOKIE_PATH,
        domain=settings.LANGUAGE_COOKIE_DOMAIN,
        secure=settings.LANGUAGE_COOKIE_SECURE,
        httponly=settings.LANGUAGE_COOKIE_HTTPONLY,
        samesite=settings.LANGUAGE_COOKIE_SAMESITE,
    )
    cookie_parts += [f"{k.replace('_', '-')}={v}" for k, v in args.items() if v]
    return "; ".join(cookie_parts)



"""Helpers for SASS/SCSS compilation."""

from colour import web2hex
from sass import SassColor
from sass_processor.storage import SassFileStorage

from .core_helpers import get_site_preferences


def get_colour(html_colour: str) -> SassColor:
    """Get a SASS colour object from an HTML colour string."""
    rgb = web2hex(html_colour, force_long=True)[1:]
    r, g, b = int(rgb[0:2], 16), int(rgb[2:4], 16), int(rgb[4:6], 16)

    return SassColor(r, g, b, 255)


def get_preference(section: str, name: str) -> str:
    """Get a preference from dynamic-preferences."""
    return get_site_preferences()[f"{section}__{name}"]


def clean_scss(*args, **kwargs) -> None:
    """Unlink compiled CSS (i.e. cache invalidation)."""
    sass_storage = SassFileStorage()
    __, files = sass_storage.listdir("public")

    for source_map in filter(lambda x: x.endswith(".css.map"), files):
        sass_storage.delete(f"public/{source_map}")



import json

import pytest

from aleksis.core.templatetags.data_helpers import get_dict, parse_json, verbose_name

pytestmark = pytest.mark.django_db


def test_get_dict_object():
    class _Foo(object):
        bar = 12

    assert _Foo.bar == get_dict(_Foo, "bar")


def test_get_dict_dict():
    _foo = {"bar": 12}

    assert _foo["bar"] == get_dict(_foo, "bar")


def test_get_dict_list():
    _foo = [10, 11, 12]

    assert _foo[2] == get_dict(_foo, 2)


def test_get_dict_invalid():
    _foo = 12

    assert get_dict(_foo, "bar") is None


def test_verbose_name_model():
    assert verbose_name("core", "person") == "Person"


def test_verbose_name_field():
    assert verbose_name("core", "person", "first_name") == "First Name"


def test_parse_json_json():
    foo = {"foo": 12, "bar": "12", "baz": []}
    foo_json = json.dumps(foo)

    assert parse_json(foo_json) == foo
    assert parse_json("{}") == {}


def test_parse_json_empty():
    assert parse_json(None) is None
    assert parse_json("") is None



# Generated by Django 3.1.3 on 2020-11-14 16:11

import django.contrib.sites.managers
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ("sites", "0002_alter_domain_unique"),
        ("contenttypes", "0002_remove_content_type_name"),
        ("core", "0007_dashboard_widget_order"),
    ]

    operations = [
        migrations.CreateModel(
            name="DataCheckResult",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("extended_data", models.JSONField(default=dict, editable=False)),
                (
                    "check",
                    models.CharField(
                        choices=[], max_length=255, verbose_name="Related data check task"
                    ),
                ),
                ("object_id", models.CharField(max_length=255)),
                ("solved", models.BooleanField(default=False, verbose_name="Issue solved")),
                ("sent", models.BooleanField(default=False, verbose_name="Notification sent")),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="contenttypes.contenttype"
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        default=1,
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="sites.site",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("run_data_checks", "Can run data checks"),
                    ("solve_data_problem", "Can solve data check problems"),
                ),
                "verbose_name": "Data check result",
                "verbose_name_plural": "Data check results",
            },
            managers=[("objects", django.contrib.sites.managers.CurrentSiteManager()),],
        ),
    ]



from textwrap import wrap

from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy as _

import django_tables2 as tables
from django_tables2.utils import A

from .models import Person
from .util.core_helpers import get_site_preferences


class SchoolTermTable(tables.Table):
    """Table to list persons."""

    class Meta:
        attrs = {"class": "highlight"}

    name = tables.LinkColumn("edit_school_term", args=[A("id")])
    date_start = tables.Column()
    date_end = tables.Column()
    edit = tables.LinkColumn(
        "edit_school_term",
        args=[A("id")],
        text=_("Edit"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange orange-text"}},
        verbose_name=_("Actions"),
    )


class PersonsTable(tables.Table):
    """Table to list persons."""

    class Meta:
        model = Person
        attrs = {"class": "highlight"}
        fields = []

    first_name = tables.LinkColumn("person_by_id", args=[A("id")])
    last_name = tables.LinkColumn("person_by_id", args=[A("id")])


class FullPersonsTable(PersonsTable):
    """Table to list persons."""

    photo = tables.Column(verbose_name=_("Photo"), accessor="pk", orderable=False)
    address = tables.Column(verbose_name=_("Address"), accessor="pk", orderable=False)

    class Meta(PersonsTable.Meta):
        fields = (
            "photo",
            "short_name",
            "date_of_birth",
            "sex",
            "email",
            "user__username",
        )
        sequence = ("photo", "first_name", "last_name", "short_name", "...")

    def render_photo(self, value, record):
        return render_to_string(
            "core/partials/avatar_content.html",
            {
                "person_or_user": record,
                "class": "materialize-circle table-circle",
                "img_class": "materialize-circle",
            },
            self.request,
        )

    def render_address(self, value, record):
        return render_to_string(
            "core/partials/address.html",
            {
                "person": record,
            },
            self.request,
        )

    def before_render(self, request):
        """Hide columns if user has no permission to view them."""
        if not self.request.user.has_perm("core.view_person_rule"):
            self.columns.hide("date_of_birth")
            self.columns.hide("sex")
        if not self.request.user.has_perm("core.view_contact_details_rule"):
            self.columns.hide("email")
        if not self.request.user.has_perm("core.view_address"):
            self.columns.hide("street")
            self.columns.hide("housenumber")
            self.columns.hide("postal_code")
            self.columns.hide("place")


class GroupsTable(tables.Table):
    """Table to list groups."""

    class Meta:
        attrs = {"class": "highlight"}

    name = tables.LinkColumn("group_by_id", args=[A("id")])
    short_name = tables.LinkColumn("group_by_id", args=[A("id")])
    school_term = tables.Column()


class AdditionalFieldsTable(tables.Table):
    """Table to list group types."""

    class Meta:
        attrs = {"class": "hightlight"}

    title = tables.LinkColumn("edit_additional_field_by_id", args=[A("id")])
    delete = tables.LinkColumn(
        "delete_additional_field_by_id",
        args=[A("id")],
        verbose_name=_("Delete"),
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red"}},
    )


class GroupTypesTable(tables.Table):
    """Table to list group types."""

    class Meta:
        attrs = {"class": "highlight"}

    name = tables.LinkColumn("edit_group_type_by_id", args=[A("id")])
    description = tables.LinkColumn("edit_group_type_by_id", args=[A("id")])
    delete = tables.LinkColumn(
        "delete_group_type_by_id", args=[A("id")], verbose_name=_("Delete"), text=_("Delete")
    )


class DashboardWidgetTable(tables.Table):
    """Table to list dashboard widgets."""

    class Meta:
        attrs = {"class": "highlight"}

    widget_name = tables.Column(accessor="pk")
    title = tables.LinkColumn("edit_dashboard_widget", args=[A("id")])
    active = tables.BooleanColumn(yesno="check,cancel", attrs={"span": {"class": "material-icons"}})
    delete = tables.LinkColumn(
        "delete_dashboard_widget",
        args=[A("id")],
        text=_("Delete"),
        attrs={"a": {"class": "btn-flat waves-effect waves-red red-text"}},
        verbose_name=_("Actions"),
    )

    def render_widget_name(self, value, record):
        return record._meta.verbose_name


class PersonColumn(tables.Column):
    """Returns person object from given id."""

    def render(self, value):
        return Person.objects.get(pk=value)


class InvitationCodeColumn(tables.Column):
    """Returns invitation code in a more readable format."""

    def render(self, value):
        packet_size = get_site_preferences()["auth__invite_code_packet_size"]
        return "-".join(wrap(value, packet_size))


class InvitationsTable(tables.Table):
    """Table to list persons."""

    person_id = PersonColumn()
    email = tables.EmailColumn()
    sent = tables.DateColumn()
    inviter_id = PersonColumn()
    key = InvitationCodeColumn()
    accepted = tables.BooleanColumn(
        yesno="check,cancel", attrs={"span": {"class": "material-icons"}}
    )


class PermissionDeleteColumn(tables.LinkColumn):
    """Link column with label 'Delete'."""

    def __init__(self, url, **kwargs):
        super().__init__(
            url,
            args=[A("pk")],
            text=_("Delete"),
            attrs={"a": {"class": "btn-flat waves-effect waves-red red-text"}},
            verbose_name=_("Actions"),
            **kwargs
        )


class PermissionTable(tables.Table):
    """Table to list permissions."""

    class Meta:
        attrs = {"class": "responsive-table highlight"}

    permission = tables.Column()


class ObjectPermissionTable(PermissionTable):
    """Table to list object permissions."""

    content_object = tables.Column()


class GlobalPermissionTable(PermissionTable):
    """Table to list global permissions."""

    pass


class GroupObjectPermissionTable(ObjectPermissionTable):
    """Table to list assigned group object permissions."""

    group = tables.Column()
    delete = PermissionDeleteColumn("delete_group_object_permission")


class UserObjectPermissionTable(ObjectPermissionTable):
    """Table to list assigned user object permissions."""

    user = tables.Column()
    delete = PermissionDeleteColumn("delete_user_object_permission")


class GroupGlobalPermissionTable(GlobalPermissionTable):
    """Table to list assigned global user permissions."""

    group = tables.Column()
    delete = PermissionDeleteColumn("delete_group_global_permission")


class UserGlobalPermissionTable(GlobalPermissionTable):
    """Table to list assigned global group permissions."""

    user = tables.Column()
    delete = PermissionDeleteColumn("delete_user_global_permission")



from django.http import HttpResponse
from django.test import override_settings
from django.urls import path, reverse
from django.views.generic import View

from oauth2_provider.views.mixins import ScopedResourceMixin

from aleksis.core.util.auth_helpers import ClientProtectedResourceMixin


class TestViewClientProtectedResourceMixin(ScopedResourceMixin, ClientProtectedResourceMixin, View):
    required_scopes = ["write"]

    def get(self, request):
        return HttpResponse("OK")


urlpatterns = [
    path(
        "client_protected_resource_mixin_test/",
        TestViewClientProtectedResourceMixin.as_view(),
        name="client_protected_resource_mixin_test",
    ),
]



"""Utilities and extensions for django_auth_ldap."""

from django.core.exceptions import PermissionDenied

from django_auth_ldap.backend import LDAPBackend as _LDAPBackend

from ..models import UserAdditionalAttributes


class LDAPBackend(_LDAPBackend):
    default_settings = {"SET_USABLE_PASSWORD": False}

    def authenticate_ldap_user(self, ldap_user, password):
        """Authenticate user against LDAP and set local password if successful.

        Having a local password is needed to make changing passwords easier. In
        order to catch password changes in a universal way and forward them to
        backends (like LDAP, in this case), getting the old password first is
        necessary to authenticate as that user to LDAP.

        We buy the small insecurity of having a hash of the password in the
        Django database in order to not require it to have global admin permissions
        on the LDAP directory.
        """
        user = super().authenticate_ldap_user(ldap_user, password)

        if self.settings.SET_USABLE_PASSWORD:
            if not user:
                # The user could not be authenticated against LDAP.
                # We need to make sure to let other backends handle it, but also that
                # we do not let actually deleted/locked LDAP users fall through to a
                # backend that cached a valid password
                if UserAdditionalAttributes.get_user_attribute(
                    ldap_user._username, "ldap_authenticated", False
                ):
                    # User was LDAP-authenticated in the past, so we fail authentication now
                    # to not let other backends override a legitimate deletion
                    raise PermissionDenied("LDAP failed to authenticate user")
                else:
                    # No note about LDAP authentication in the past
                    # The user can continue authentication like before if they exist
                    return user

            # Set a usable password so users can change their LDAP password
            user.set_password(password)
            user.save()

            # Not that we LDAP-autenticated the user so we can check this in the future
            UserAdditionalAttributes.set_user_attribute(
                ldap_user._username, "ldap_authenticated", True
            )

        return user



# Generated by Django 3.1.4 on 2021-02-19 13:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0016_taskuserassignment'),
    ]

    operations = [
        migrations.AddField(
            model_name='dashboardwidget',
            name='broken',
            field=models.BooleanField(default=False, verbose_name='Widget is broken'),
        ),
    ]



from django.conf import settings
from django.utils.translation import gettext_lazy as _

MENUS = {
    "NAV_MENU_CORE": [
        {
            "name": _("Login"),
            "url": settings.LOGIN_URL,
            "svg_icon": "mdi:login-variant",
            "vuetify_icon": "mdi-login-variant",
            "validators": ["menu_generator.validators.is_anonymous"],
        },
        {
            "name": _("Sign up"),
            "url": "account_signup",
            "svg_icon": "mdi:account-plus-outline",
            "vuetify_icon": "mdi-account-plus-outline",
            "validators": [
                "menu_generator.validators.is_anonymous",
                ("aleksis.core.util.predicates.permission_validator", "core.can_register"),
            ],
        },
        {
            "name": _("Accept invitation"),
            "url": "enter_invitation_code",
            "svg_icon": "mdi:key-outline",
            "vuetify_icon": "mdi-key-outline",
            "validators": [
                "menu_generator.validators.is_anonymous",
                ("aleksis.core.util.predicates.permission_validator", "core.invite_enabled"),
            ],
        },
        {
            "name": _("Dashboard"),
            "url": "index",
            "svg_icon": "mdi:home-outline",
            "vuetify_icon": "mdi-home-outline",
            "validators": [
                ("aleksis.core.util.predicates.permission_validator", "core.view_dashboard_rule")
            ],
        },
        {
            "name": _("Admin"),
            "url": "#",
            "svg_icon": "mdi:security",
            "vuetify_icon": "mdi-security",
            "validators": [
                ("aleksis.core.util.predicates.permission_validator", "core.view_admin_menu"),
            ],
            "submenu": [
                {
                    "name": _("Announcements"),
                    "url": "announcements",
                    "svg_icon": "mdi:message-alert-outline",
                    "vuetify_icon": "mdi-message-alert-outline",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_announcements_rule",
                        ),
                    ],
                },
                {
                    "name": _("School terms"),
                    "url": "school_terms",
                    "svg_icon": "mdi:calendar-range-outline",
                    "vuetify_icon": "mdi-calendar-range-outline",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_schoolterm_rule",
                        ),
                    ],
                },
                {
                    "name": _("Dashboard widgets"),
                    "url": "dashboard_widgets",
                    "svg_icon": "mdi:view-dashboard-outline",
                    "vuetify_icon": "mdi-view-dashboard-outline",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_dashboardwidget_rule",
                        ),
                    ],
                },
                {
                    "name": _("Data management"),
                    "url": "data_management",
                    "svg_icon": "mdi:chart-donut",
                    "vuetify_icon": "mdi-chart-donut",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.manage_data_rule",
                        ),
                    ],
                },
                {
                    "name": _("System status"),
                    "url": "system_status",
                    "svg_icon": "mdi:power-settings",
                    "vuetify_icon": "mdi-power-settings",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_system_status_rule",
                        ),
                    ],
                },
                {
                    "name": _("Configuration"),
                    "url": "preferences_site",
                    "svg_icon": "mdi:tune",
                    "vuetify_icon": "mdi-tune",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.change_site_preferences_rule",
                        ),
                    ],
                },
                {
                    "name": _("Data checks"),
                    "url": "check_data",
                    "svg_icon": "mdi:list-status",
                    "vuetify_icon": "mdi-list-status",
                    "validators": ["menu_generator.validators.is_superuser"],
                },
                {
                    "name": _("Manage permissions"),
                    "url": "manage_user_global_permissions",
                    "svg_icon": "mdi:shield-outline",
                    "vuetify_icon": "mdi-shield-outline",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.manage_permissions",
                        ),
                    ],
                },
                {
                    "name": _("Backend Admin"),
                    "url": "admin:index",
                    "svg_icon": "mdi:database-cog-outline",
                    "vuetify_icon": "mdi-database-cog-outline",
                    "validators": [
                        "menu_generator.validators.is_superuser",
                    ],
                },
                {
                    "name": _("OAuth2 Applications"),
                    "url": "oauth2_applications",
                    "svg_icon": "mdi:gesture-tap-hold",
                    "vuetify_icon": "mdi-gesture-tap-hold",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.list_oauth_applications_rule",
                        ),
                    ],
                },
            ],
        },
        {
            "name": _("People"),
            "url": "#",
            "svg_icon": "mdi:account-group-outline",
            "vuetify_icon": "mdi-account-group-outline",
            "root": True,
            "validators": [
                ("aleksis.core.util.predicates.permission_validator", "core.view_people_menu_rule")
            ],
            "submenu": [
                {
                    "name": _("Persons"),
                    "url": "persons",
                    "svg_icon": "mdi:account-outline",
                    "vuetify_icon": "mdi-account-outline",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_persons_rule",
                        )
                    ],
                },
                {
                    "name": _("Groups"),
                    "url": "groups",
                    "svg_icon": "mdi:account-multiple-outline",
                    "vuetify_icon": "mdi-account-multiple-outline",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_groups_rule",
                        )
                    ],
                },
                {
                    "name": _("Group types"),
                    "url": "group_types",
                    "svg_icon": "mdi:shape-outline",
                    "vuetify_icon": "mdi-shape-outline",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_grouptypes_rule",
                        )
                    ],
                },
                {
                    "name": _("Groups and child groups"),
                    "url": "groups_child_groups",
                    "svg_icon": "mdi:account-multiple-plus-outline",
                    "vuetify_icon": "mdi-account-multiple-plus-outline",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.assign_child_groups_to_groups_rule",
                        )
                    ],
                },
                {
                    "name": _("Additional fields"),
                    "url": "additional_fields",
                    "svg_icon": "mdi:palette-swatch-outline",
                    "vuetify_icon": "mdi-palette-swatch-outline",
                    "validators": [
                        (
                            "aleksis.core.util.predicates.permission_validator",
                            "core.view_additionalfields_rule",
                        )
                    ],
                },
                {
                    "name": _("Invite person"),
                    "url": "invite_person",
                    "svg_icon": "mdi:account-plus-outline",
                    "vuetify_icon": "mdi-account-plus-outline",
                    "validators": [
                        "menu_generator.validators.is_authenticated",
                        ("aleksis.core.util.predicates.permission_validator", "core.can_invite"),
                    ],
                },
            ],
        },
    ],
    "DATA_MANAGEMENT_MENU": [
        {
            "name": _("Assign child groups to groups"),
            "url": "groups_child_groups",
            "validators": [
                (
                    "aleksis.core.util.predicates.permission_validator",
                    "core.assign_child_groups_to_groups_rule",
                )
            ],
        },
    ],
    "NAVBAR_ACCOUNT_MENU": [
        {
            "name": _("Stop impersonation"),
            "url": "impersonate-stop",
            "svg_icon": "mdi:stop",
            "vuetify_icon": "mdi-stop",
            "validators": [
                "menu_generator.validators.is_authenticated",
                "aleksis.core.util.core_helpers.is_impersonate",
            ],
        },
        {
            "name": _("Account"),
            "url": "person",
            "svg_icon": "mdi:account-outline",
            "vuetify_icon": "mdi-account-outline",
            "validators": [
                "menu_generator.validators.is_authenticated",
                "aleksis.core.util.core_helpers.has_person",
            ],
        },
        {
            "name": _("Preferences"),
            "url": "preferences_person",
            "svg_icon": "mdi:cog-outline",
            "vuetify_icon": "mdi-cog-outline",
            "validators": [
                "menu_generator.validators.is_authenticated",
                "aleksis.core.util.core_helpers.has_person",
            ],
        },
        {
            "name": _("2FA"),
            "url": "two_factor:profile",
            "svg_icon": "mdi:two-factor-authentication",
            "vuetify_icon": "mdi-two-factor-authentication",
            "validators": [
                "menu_generator.validators.is_authenticated",
            ],
        },
        {
            "name": _("Change password"),
            "url": "account_change_password",
            "svg_icon": "mdi:form-textbox-password",
            "vuetify_icon": "mdi-form-textbox-password",
            "validators": [
                "menu_generator.validators.is_authenticated",
                (
                    "aleksis.core.util.predicates.permission_validator",
                    "core.can_change_password",
                ),
            ],
        },
        {
            "name": _("Third-party accounts"),
            "url": "socialaccount_connections",
            "svg_icon": "mdi:earth",
            "vuetify_icon": "mdi-earth",
            "validators": [
                "menu_generator.validators.is_authenticated",
                "aleksis.core.util.core_helpers.has_person",
            ],
        },
        {
            "name": _("Authorized applications"),
            "url": "oauth2_provider:authorized-token-list",
            "svg_icon": "mdi:gesture-tap-hold",
            "vuetify_icon": "mdi-gesture-tap-hold",
            "validators": [
                "menu_generator.validators.is_authenticated",
                "aleksis.core.util.core_helpers.has_person",
            ],
        },
        {
            "name": _("Calendar Feeds"),
            "url": "ical_feed_list",
            "svg_icon": "mdi:calendar-multiple",
            "vuetify_icon": "mdi-calendar-multiple",
            "validators": [
                "menu_generator.validators.is_authenticated",
                (
                    "aleksis.core.util.predicates.permission_validator",
                    "core.view_ical_rule",
                ),
            ],
        },
        {
            "divider": True,
            "name": _("Logout"),
            "url": "logout",
            "svg_icon": "mdi:logout-variant",
            "vuetify_icon": "mdi-logout-variant",
            "validators": ["menu_generator.validators.is_authenticated"],
        },
    ],
}



# Generated by Django 3.0.7 on 2020-06-28 14:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0003_drop_image_cropping'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='group',
            options={'ordering': ['short_name', 'name'], 'permissions': (('assign_child_groups_to_groups', 'Can assign child groups to groups'), ('view_group_stats', 'Can view statistics about group.')), 'verbose_name': 'Group', 'verbose_name_plural': 'Groups'},
        ),
        migrations.AlterField(
            model_name='group',
            name='additional_fields',
            field=models.ManyToManyField(blank=True, to='core.AdditionalField', verbose_name='Additional fields'),
        ),
    ]



from datetime import timedelta
from time import sleep
from unittest.mock import patch

from django.test import override_settings
from django.utils import timezone

import pytest
from freezegun import freeze_time

from aleksis.core.models import Notification, Person
from aleksis.core.util.notifications import _send_due_notifications

pytestmark = pytest.mark.django_db


def test_send_notification():
    email = "doe@example.com"
    recipient = Person.objects.create(first_name="Jane", last_name="Doe", email=email)

    sender = "Foo"
    title = "There is happened something."
    description = "Here you get some more information."
    link = "https://aleksis.org/"

    notification = Notification(
        sender=sender,
        recipient=recipient,
        title=title,
        description=description,
        link=link,
    )

    with patch("aleksis.core.models.Notification.send") as patched_send:
        patched_send.assert_not_called()

        notification.save()

        patched_send.assert_called()


def test_send_scheduled_notification():
    email = "doe@example.com"
    recipient = Person.objects.create(first_name="Jane", last_name="Doe", email=email)

    sender = "Foo"
    title = "There is happened something."
    description = "Here you get some more information."
    link = "https://aleksis.org/"

    notification = Notification(
        sender=sender,
        recipient=recipient,
        title=title,
        description=description,
        link=link,
        send_at=timezone.now() + timedelta(days=1),
    )
    notification.save()

    with patch("aleksis.core.models.Notification.send") as patched_send:
        patched_send.assert_not_called()

        _send_due_notifications()

        patched_send.assert_not_called()

        with freeze_time(timezone.now() + timedelta(days=1)):
            _send_due_notifications()

        patched_send.assert_called()


@override_settings(CELERY_BROKER_URL="memory://localhost//")
@pytest.mark.django_db(
    databases=["default", "default_oot"], serialized_rollback=True, transaction=True
)
def test_email_notification(mailoutbox, celery_worker):
    email = "doe@example.com"
    recipient = Person.objects.create(first_name="Jane", last_name="Doe", email=email)

    sender = "Foo"
    title = "There is happened something."
    description = "Here you get some more information."
    link = "https://aleksis.org/"

    notification = Notification(
        sender=sender, recipient=recipient, title=title, description=description, link=link
    )
    notification.save()

    sleep(3)

    notification.refresh_from_db()
    assert notification.sent

    assert len(mailoutbox) == 1

    mail = mailoutbox[0]

    assert email in mail.to
    assert title in mail.body
    assert description in mail.body
    assert link in mail.body
    assert sender in mail.body
    assert recipient.addressing_name in mail.subject
    assert recipient.addressing_name in mail.body



from django.conf import settings
from django.utils import translation

import graphene
from graphene import ObjectType
from graphene_django import DjangoObjectType

from .models import Group, Notification, Person
from .util.core_helpers import get_app_module, get_app_packages, has_person
from .util.frontend_helpers import get_language_cookie


class NotificationType(DjangoObjectType):
    class Meta:
        model = Notification
        fields = [
            "sender",
            "recipient",
            "title",
            "description",
            "link",
            "icon",
            "send_at",
            "read",
            "sent",
            "created",
            "modified",
        ]

        @staticmethod
        def resolve_recipient(root, info, **kwargs):
            if info.context.user.has_perm("core.view_person_rule", root.recipient):
                return root.recipient
            raise PermissionDenied()


class PersonType(DjangoObjectType):
    class Meta:
        model = Person
        fields = [
            "user",
            "first_name",
            "last_name",
            "additional_name",
            "short_name",
            "street",
            "housenumber",
            "postal_code",
            "place",
            "phone_number",
            "mobile_number",
            "email",
            "date_of_birth",
            "place_of_birth",
            "sex",
            "photo",
            "avatar",
            "guardians",
            "primary_group",
            "description",
            "children",
            "owner_of",
            "member_of",
        ]

    full_name = graphene.Field(graphene.String)

    def resolve_full_name(root: Person, info, **kwargs):
        return root.full_name

    def resolve_street(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_address_rule", root):
            return root.street
        return None

    def resolve_housenumber(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_address_rule", root):
            return root.housenumber
        return None

    def resolve_postal_code(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_address_rule", root):
            return root.postal_code
        return None

    def resolve_place(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_address_rule", root):
            return root.place
        return None

    def resolve_phone_number(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_contact_details_rule", root):
            return root.phone_number
        return None

    def resolve_mobile_number(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_contact_details_rule", root):
            return root.mobile_number
        return None

    def resolve_email(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_contact_details_rule", root):
            return root.email
        return None

    def resolve_date_of_birth(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_personal_details_rule", root):
            return root.date_of_birth
        return None

    def resolve_place_of_birth(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_personal_details_rule", root):
            return root.place_of_birth
        return None

    def resolve_children(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_personal_details_rule", root):
            return get_objects_for_user(info.context.user, "core.view_person", root.children.all())
        return []

    def resolve_guardians(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_personal_details_rule", root):
            return get_objects_for_user(info.context.user, "core.view_person", root.guardians.all())
        return []

    def resolve_member_of(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_person_groups_rule", root):
            return get_objects_for_user(info.context.user, "core.view_group", root.member_of.all())
        return []

    def resolve_owner_of(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_person_groups_rule", root):
            return get_objects_for_user(info.context.user, "core.view_group", root.owner_of.all())
        return []

    def resolve_primary_group(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_group_rule", root.primary_group):
            return root.primary_group
        raise PermissionDenied()

    def resolve_photo(root, info, **kwargs):
        if info.context.user.has_perm("core.view_photo_rule", root):
            return root.photo
        return None

    def resolve_avatar(root, info, **kwargs):
        if info.context.user.has_perm("core.view_avatar_rule", root):
            return root.avatar
        return None


class GroupType(DjangoObjectType):
    class Meta:
        model = Group
        fields = [
            "name",
            "short_name",
            "members",
            "owners",
            "parent_groups",
            "group_type",
            "additional_fields",
            "photo",
            "avatar",
        ]

    @staticmethod
    def resolve_parent_groups(root, info, **kwargs):
        return get_objects_for_user(info.context.user, "core.view_group", root.parent_groups.all())

    @staticmethod
    def resolve_members(root, info, **kwargs):
        persons = get_objects_for_user(info.context.user, "core.view_person", root.members.all())
        if has_person(info.context.user) and [
            m for m in root.members.all() if m.pk == info.context.user.person.pk
        ]:
            persons = (persons | Person.objects.get(pk=info.context.user.person.pk)).distinct()
        return persons

    @staticmethod
    def resolve_owners(root, info, **kwargs):
        persons = get_objects_for_user(info.context.user, "core.view_person", root.owners.all())
        if has_person(info.context.user) and [
            o for o in root.owners.all() if o.pk == info.context.user.person.pk
        ]:
            persons = (persons | Person.objects.get(pk=info.context.user.person.pk)).distinct()
        return persons

    @staticmethod
    def resolve_group_type(root, info, **kwargs):
        if info.context.user.has_perm("core.view_grouptype_rule", root.group_type):
            return root.group_type
        raise PermissionDenied()

    @staticmethod
    def resolve_additional_fields(root, info, **kwargs):
        return get_objects_for_user(
            info.context.user, "core.view_additionalfield", root.additional_fields.all()
        )


class LanguageType(ObjectType):
    code = graphene.String(required=True)
    name = graphene.String(required=True)
    name_local = graphene.String(required=True)
    name_translated = graphene.String(required=True)
    bidi = graphene.Boolean(required=True)
    cookie = graphene.String(required=True)


class SystemPropertiesType(graphene.ObjectType):
    current_language = graphene.String(required=True)
    available_languages = graphene.List(LanguageType)

    def resolve_current_language(parent, info, **kwargs):
        return info.context.LANGUAGE_CODE

    def resolve_available_languages(parent, info, **kwargs):
        return [
            translation.get_language_info(code) | {"cookie": get_language_cookie(code)}
            for code, name in settings.LANGUAGES
        ]


class MarkNotificationReadMutation(graphene.Mutation):
    class Arguments:
        id = graphene.ID()  # noqa

    notification = graphene.Field(NotificationType)

    @classmethod
    def mutate(cls, root, info, id):  # noqa
        notification = Notification.objects.get(pk=id, recipient=info.context.user.person)

        notification.read = True
        notification.save()

        return notification


class Query(graphene.ObjectType):
    ping = graphene.String(default_value="pong")

    who_am_i = graphene.Field(PersonType)

    system_properties = graphene.Field(SystemPropertiesType)

    def resolve_who_am_i(root, info, **kwargs):
        if has_person(info.context.user):
            return info.context.user.person
        else:
            return None

    def resolve_system_properties(root, info, **kwargs):
        return True


class Mutation(graphene.ObjectType):
    mark_notification_read = MarkNotificationReadMutation.Field()


def build_global_schema():
    """Build global GraphQL schema from all apps."""
    query_bases = [Query]
    mutation_bases = [Mutation]

    for app in get_app_packages():
        schema_mod = get_app_module(app, "schema")
        if not schema_mod:
            # The app does not define a schema
            continue

        if AppQuery := getattr(schema_mod, "Query", None):
            query_bases.append(AppQuery)
        if AppMutation := getattr(schema_mod, "Mutation", None):
            mutation_bases.append(AppMutation)

    # Define classes using all query/mutation classes as mixins
    #  cf. https://docs.graphene-python.org/projects/django/en/latest/schema/#adding-to-the-schema
    GlobalQuery = type("GlobalQuery", tuple(query_bases), {})
    GlobalMutation = type("GlobalMutation", tuple(mutation_bases), {})

    return graphene.Schema(query=GlobalQuery, mutation=GlobalMutation)


schema = build_global_schema()



# Generated by Django 3.2.16 on 2022-11-03 11:36

from django.db import migrations, models
import django.utils.timezone
import oauth2_provider.generators
import oauth2_provider.models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0041_update_gender_choices'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pdffile',
            name='html_file',
            field=models.FileField(blank=True, null=True, upload_to='pdfs/', verbose_name='Generated HTML file'),
        ),
    ]



from datetime import datetime

from django.conf import settings
from django.utils.translation import gettext as _

from dbbackup import utils as dbbackup_utils
from dbbackup.storage import get_storage
from django_celery_results.models import TaskResult
from health_check.backends import BaseHealthCheckBackend

from aleksis.core.models import DataCheckResult


class DataChecksHealthCheckBackend(BaseHealthCheckBackend):
    """Checks whether there are unresolved data problems."""

    critical_service = False

    def check_status(self):
        if DataCheckResult.objects.filter(solved=False).exists():
            self.add_error(_("There are unresolved data problems."))

    def identifier(self):
        return self.__class__.__name__


class BaseBackupHealthCheck(BaseHealthCheckBackend):
    """Common base class for backup age checks."""

    critical_service = False
    content_type = None
    configured_seconds = None

    def check_status(self):
        storage = get_storage()

        try:
            backups = storage.list_backups(content_type=self.content_type)
        except Exception as ex:
            self.add_error(_("Error accessing backup storage: {}").format(str(ex)))
            return

        if backups:
            last_backup = backups[:1]
            last_backup_time = dbbackup_utils.filename_to_date(last_backup[0])
            time_gone_since_backup = datetime.now() - last_backup_time

            # Check if backup is older than configured time
            if time_gone_since_backup.seconds > self.configured_seconds:
                self.add_error(_("Last backup {}!").format(time_gone_since_backup))
        else:
            self.add_error(_("No backup found!"))


class DbBackupAgeHealthCheck(BaseBackupHealthCheck):
    """Checks if last backup file is less than configured seconds ago."""

    content_type = "db"
    configured_seconds = settings.DBBACKUP_CHECK_SECONDS


class MediaBackupAgeHealthCheck(BaseBackupHealthCheck):
    """Checks if last backup file is less than configured seconds ago."""

    content_type = "media"
    configured_seconds = settings.MEDIABACKUP_CHECK_SECONDS


class BackupJobHealthCheck(BaseHealthCheckBackend):
    """Checks if last backup file is less than configured seconds ago."""

    critical_service = False

    def check_status(self):
        task = TaskResult.objects.filter(task_name="aleksis.core.tasks.backup_data").last()

        # Check if state is success
        if not task:
            self.add_error(_("No backup result found!"))
        elif task and task.status != "SUCCESS":
            self.add_error(f"{task.status} - {task.result}")



from django.conf import settings
from django.utils.formats import date_format
from django.utils.functional import classproperty
from django.utils.translation import gettext_lazy as _

from django_ical.utils import build_rrule_from_text
from django_ical.views import ICalFeed

from aleksis.core import models
from aleksis.core.util.core_helpers import get_site_preferences, queryset_rules_filter


class PersonalICalFeedBase(ICalFeed):
    """Base class for personal iCal feeds."""

    @property
    def product_id(self):
        lang = self.request.LANGUAGE_CODE
        title = get_site_preferences()["general__title"]
        return f"-//AlekSIS//{title}//{lang}"

    link = settings.BASE_URL
    timezone = settings.TIME_ZONE
    person = None
    request = None

    def get_object(self, request, *args, **kwargs):
        if kwargs.get("person"):
            self.person = kwargs.pop("person")
        self.request = request
        return super().get_object(request, *args, **kwargs)

    @classproperty
    def subclasses_list(cls):
        return cls.__subclasses__()

    @classproperty
    def subclasses_dict(cls):
        return {subclass.__name__: subclass for subclass in cls.subclasses_list}

    @classproperty
    def subclass_choices(cls):
        return [
            (subclass.__name__, f"{subclass.title} – {subclass.description}")
            for subclass in cls.subclasses_list
        ]


class BirthdayFeed(PersonalICalFeedBase):
    """Birthday calendar feed."""

    title = _("Birthday Calendar")
    description = _("A Calendar of Birthdays")
    file_name = "birthdays.ics"

    def items(self):
        from aleksis.core.models import Person

        return queryset_rules_filter(
            obj=self.person.user,
            perm="core.view_personal_details_rule",
            queryset=Person.objects.filter(date_of_birth__isnull=False),
        )

    def item_title(self, item: "models.Person"):
        return _("%(name)s's birthday") % {
            "name": item.addressing_name,
        }

    def item_description(self, item: "models.Person"):
        return _("%(name)s was born on %(birthday)s") % {
            "name": item.addressing_name,
            "birthday": date_format(item.date_of_birth),
        }

    def item_start_datetime(self, item):
        return item.date_of_birth

    def item_rrule(self, item):
        return build_rrule_from_text("FREQ=YEARLY")



# flake8: noqa: DJ12

import os
from datetime import datetime
from typing import Any, Callable, List, Optional, Union

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.views import LoginView, SuccessURLAllowedHostsMixin
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.managers import CurrentSiteManager
from django.contrib.sites.models import Site
from django.db import models
from django.db.models import JSONField, QuerySet
from django.db.models.fields import CharField, TextField
from django.forms.forms import BaseForm
from django.forms.models import ModelForm, ModelFormMetaclass
from django.http import HttpResponse
from django.utils.functional import lazy
from django.utils.translation import gettext as _
from django.views.generic import CreateView, UpdateView
from django.views.generic.edit import DeleteView, ModelFormMixin

import reversion
from dynamic_preferences.settings import preferences_settings
from dynamic_preferences.types import FilePreference
from guardian.admin import GuardedModelAdmin
from guardian.core import ObjectPermissionChecker
from jsonstore.fields import IntegerField, JSONFieldMixin
from material.base import Layout, LayoutNode
from polymorphic.base import PolymorphicModelBase
from polymorphic.managers import PolymorphicManager
from polymorphic.models import PolymorphicModel
from rules.contrib.admin import ObjectPermissionsModelAdmin

from aleksis.core.managers import (
    CurrentSiteManagerWithoutMigrations,
    PolymorphicCurrentSiteManager,
    SchoolTermRelatedQuerySet,
)


class _ExtensibleModelBase(models.base.ModelBase):
    """Ensure predefined behaviour on model creation.

    This metaclass serves the following purposes:

     - Register all AlekSIS models with django-reverseion
    """

    def __new__(mcls, name, bases, attrs):
        mcls = super().__new__(mcls, name, bases, attrs)

        if "Meta" not in attrs or not attrs["Meta"].abstract:
            # Register all non-abstract models with django-reversion
            mcls = reversion.register(mcls)

            mcls.extra_permissions = []

        return mcls


def _generate_one_to_one_proxy_property(field, subfield):
    def getter(self):
        if hasattr(self, field.name):
            related = getattr(self, field.name)
            return getattr(related, subfield.name)
        # Related instane does not exist
        return None

    def setter(self, val):
        if hasattr(self, field.name):
            related = getattr(self, field.name)
        else:
            # Auto-create related instance (but do not save)
            related = field.related_model()
            setattr(related, field.remote_field.name, self)
            # Ensure the related model is saved later
            self._save_reverse = getattr(self, "_save_reverse", []) + [related]
        setattr(related, subfield.name, val)

    return property(getter, setter)


class ExtensibleModel(models.Model, metaclass=_ExtensibleModelBase):
    """Base model for all objects in AlekSIS apps.

    This base model ensures all objects in AlekSIS apps fulfill the
    following properties:

     * `versions` property to retrieve all versions of the model from reversion
     * Allow injection of fields and code from AlekSIS apps to extend
       model functionality.

    Injection of fields and code
    ============================

    After all apps have been loaded, the code in the `model_extensions` module
    in every app is executed. All code that shall be injected into a model goes there.

    :Example:

    .. code-block:: python

       from datetime import date, timedelta

       from jsonstore import CharField

       from aleksis.core.models import Person

       @Person.property
       def is_cool(self) -> bool:
           return True

       @Person.property
       def age(self) -> timedelta:
           return self.date_of_birth - date.today()

       Person.field(shirt_size=CharField())

    For a more advanced example, using features from the ORM, see AlekSIS-App-Chronos
    and AlekSIS-App-Alsijil.

    :Date: 2019-11-07
    :Authors:
        - Dominik George <dominik.george@teckids.org>
    """

    # Defines a material design icon associated with this type of model
    icon_ = "radio_button_unchecked"

    site = models.ForeignKey(
        Site, on_delete=models.CASCADE, default=settings.SITE_ID, editable=False
    )
    objects = CurrentSiteManager()
    objects_all_sites = models.Manager()

    extra_permissions = []

    def get_absolute_url(self) -> str:
        """Get the URL o a view representing this model instance."""
        pass

    @property
    def versions(self) -> list[tuple[str, tuple[Any, Any]]]:
        """Get all versions of this object from django-reversion.

        Includes diffs to previous version.
        """
        versions = reversion.models.Version.objects.get_for_object(self)

        versions_with_changes = []
        for i, version in enumerate(versions):
            diff = {}
            if i > 0:
                prev_version = versions[i - 1]

                for k, val in version.field_dict.items():
                    prev_val = prev_version.field_dict.get(k, None)
                    if prev_val != val:
                        diff[k] = (prev_val, val)

            versions_with_changes.append((version, diff))

        return versions_with_changes

    extended_data = JSONField(default=dict, editable=False)

    @classmethod
    def _safe_add(cls, obj: Any, name: Optional[str]) -> None:
        # Decide the name for the attribute
        if name is None:
            prop_name = obj.__name__
        else:
            if name.isidentifier():
                prop_name = name
            else:
                raise ValueError(f"{name} is not a valid name.")

        # Verify that attribute name does not clash with other names in the class
        if hasattr(cls, prop_name):
            raise ValueError(f"{prop_name} already used.")

        # Let Django's model magic add the attribute if we got here
        cls.add_to_class(name, obj)

    @classmethod
    def property_(cls, func: Callable[[], Any], name: Optional[str] = None) -> None:
        """Add the passed callable as a property."""
        cls._safe_add(property(func), name or func.__name__)

    @classmethod
    def method(cls, func: Callable[[], Any], name: Optional[str] = None) -> None:
        """Add the passed callable as a method."""
        cls._safe_add(func, name or func.__name__)

    @classmethod
    def class_method(cls, func: Callable[[], Any], name: Optional[str] = None) -> None:
        """Add the passed callable as a classmethod."""
        cls._safe_add(classmethod(func), name or func.__name__)

    @classmethod
    def field(cls, **kwargs) -> None:
        """Add the passed jsonstore field. Must be one of the fields in django-jsonstore.

        Accepts exactly one keyword argument, with the name being the desired
        model field name and the value the field instance.
        """
        # Force kwargs to be exactly one argument
        if len(kwargs) != 1:
            raise TypeError(f"field() takes 1 keyword argument but {len(kwargs)} were given")
        name, field = kwargs.popitem()

        # Force the field to be one of the jsonstore fields
        if JSONFieldMixin not in field.__class__.__mro__:
            raise TypeError("Only jsonstore fields can be added to models.")

        # Force use of the one JSONField defined in this mixin
        field.json_field_name = "extended_data"

        cls._safe_add(field, name)

    @classmethod
    def foreign_key(
        cls,
        field_name: str,
        to: models.Model,
        to_field: str = "pk",
        to_field_type: JSONFieldMixin = IntegerField,
        related_name: Optional[str] = None,
    ) -> None:
        """Add a virtual ForeignKey.

        This works by storing the primary key (or any field passed in the to_field argument)
        and adding a property that queries the desired model.

        If the foreign model also is an ExtensibleModel, a reverse mapping is also added under
        the related_name passed as argument, or this model's default related name.
        """

        id_field_name = f"{field_name}_id"
        if related_name is None:
            related_name = cls.Meta.default_related_name

        # Add field to hold key to foreign model
        id_field = to_field_type(blank=True, null=True)
        cls.field(**{id_field_name: id_field})

        @property
        def _virtual_fk(self) -> Optional[models.Model]:
            id_field_val = getattr(self, id_field_name)
            if id_field_val:
                try:
                    return to.objects.get(**{to_field: id_field_val})
                except to.DoesNotExist:
                    # We found a stale foreign key
                    setattr(self, id_field_name, None)
                    self.save()
                    return None
            else:
                return None

        @_virtual_fk.setter
        def _virtual_fk(self, value: Optional[models.Model] = None) -> None:
            if value is None:
                id_field_val = None
            else:
                id_field_val = getattr(value, to_field)
            setattr(self, id_field_name, id_field_val)

        # Add property to wrap get/set on foreign model instance
        cls._safe_add(_virtual_fk, field_name)

        # Add related property on foreign model instance if it provides such an interface
        if hasattr(to, "_safe_add"):

            def _virtual_related(self) -> models.QuerySet:
                id_field_val = getattr(self, to_field)
                return cls.objects.filter(**{id_field_name: id_field_val})

            to.property_(_virtual_related, related_name)

    @classmethod
    def get_filter_fields(cls) -> List[str]:
        """Get names of all text-searchable fields of this model."""
        fields = []
        for field in cls.syncable_fields():
            if isinstance(field, (CharField, TextField)):
                fields.append(field.name)
        return fields

    @classmethod
    def syncable_fields(
        cls, recursive: bool = True, exclude_remotes: list = []
    ) -> list[models.Field]:
        """Collect all fields that can be synced on a model.

        If recursive is True, it recurses into related models and generates virtual
        proxy fields to access fields in related models."""
        fields = []
        for field in cls._meta.get_fields():
            if field.is_relation and field.one_to_one and recursive:
                if ExtensibleModel not in field.related_model.__mro__:
                    # Related model is not extensible and thus has no syncable fields
                    continue
                if field.related_model in exclude_remotes:
                    # Remote is excluded, probably to avoid recursion
                    continue

                # Recurse into related model to get its fields as well
                for subfield in field.related_model.syncable_fields(
                    recursive, exclude_remotes + [cls]
                ):
                    # generate virtual field names for proxy access
                    name = f"_{field.name}__{subfield.name}"
                    verbose_name = f"{field.name} ({field.related_model._meta.verbose_name}) → {subfield.verbose_name}"

                    if not hasattr(cls, name):
                        # Add proxy properties to handle access to related model
                        setattr(cls, name, _generate_one_to_one_proxy_property(field, subfield))

                    # Generate a fake field class with enough API to detect attribute names
                    fields.append(
                        type(
                            "FakeRelatedProxyField",
                            (),
                            {
                                "name": name,
                                "verbose_name": verbose_name,
                                "to_python": lambda v: subfield.to_python(v),
                            },
                        )
                    )
            elif field.editable and not field.auto_created:
                fields.append(field)

        return fields

    @classmethod
    def syncable_fields_choices(cls) -> tuple[tuple[str, str]]:
        """Collect all fields that can be synced on a model."""
        return tuple(
            [(field.name, field.verbose_name or field.name) for field in cls.syncable_fields()]
        )

    @classmethod
    def syncable_fields_choices_lazy(cls) -> Callable[[], tuple[tuple[str, str]]]:
        """Collect all fields that can be synced on a model."""
        return lazy(cls.syncable_fields_choices, tuple)

    @classmethod
    def add_permission(cls, name: str, verbose_name: str):
        """Dynamically add a new permission to a model."""
        cls.extra_permissions.append((name, verbose_name))

    def set_object_permission_checker(self, checker: ObjectPermissionChecker):
        """Annotate a ``ObjectPermissionChecker`` for use with permission system."""
        self._permission_checker = checker

    def save(self, *args, **kwargs):
        """Ensure all functionality of our extensions that needs saving gets it."""
        # For auto-created remote syncable fields
        if hasattr(self, "_save_reverse"):
            for related in self._save_reverse:
                related.save()
            del self._save_reverse

        super().save(*args, **kwargs)

    class Meta:
        abstract = True


class _ExtensiblePolymorphicModelBase(_ExtensibleModelBase, PolymorphicModelBase):
    """Base class for extensible, polymorphic models."""


class ExtensiblePolymorphicModel(
    ExtensibleModel, PolymorphicModel, metaclass=_ExtensiblePolymorphicModelBase
):
    """Model class for extensible, polymorphic models."""

    objects = PolymorphicCurrentSiteManager()
    objects_all_sites = PolymorphicManager()

    class Meta:
        abstract = True


class PureDjangoModel(object):
    """No-op mixin to mark a model as deliberately not using ExtensibleModel."""

    pass


class GlobalPermissionModel(models.Model):
    """Base model for global permissions.

    This base model ensures that global permissions are not managed."""

    class Meta:
        default_permissions = ()
        abstract = True
        managed = False


class _ExtensibleFormMetaclass(ModelFormMetaclass):
    def __new__(cls, name, bases, dct):
        x = super().__new__(cls, name, bases, dct)

        # Enforce a default for the base layout for forms that o not specify one
        if hasattr(x, "layout"):
            base_layout = x.layout.elements
        else:
            base_layout = []

        x.base_layout = base_layout
        x.layout = Layout(*base_layout)

        return x


class ExtensibleForm(ModelForm, metaclass=_ExtensibleFormMetaclass):
    """Base model for extensible forms.

    This mixin adds functionality which allows
    - apps to add layout nodes to the layout used by django-material

    :Add layout nodes:

    .. code-block:: python

        from material import Fieldset

        from aleksis.core.forms import ExampleForm

        node = Fieldset("field_name")
        ExampleForm.add_node_to_layout(node)

    """

    @classmethod
    def add_node_to_layout(cls, node: Union[LayoutNode, str]):
        """Add a node to `layout` attribute.

        :param node: django-material layout node (Fieldset, Row etc.)
        :type node: LayoutNode
        """
        cls.base_layout.append(node)
        cls.layout = Layout(*cls.base_layout)


class BaseModelAdmin(GuardedModelAdmin, ObjectPermissionsModelAdmin):
    """A base class for ModelAdmin combining django-guardian and rules."""

    pass


class SuccessMessageMixin(ModelFormMixin):
    success_message: Optional[str] = None

    def form_valid(self, form: BaseForm) -> HttpResponse:
        if self.success_message:
            messages.success(self.request, self.success_message)
        return super().form_valid(form)


class SuccessNextMixin(SuccessURLAllowedHostsMixin):
    redirect_field_name = "next"

    def get_success_url(self) -> str:
        return LoginView.get_redirect_url(self) or super().get_success_url()


class AdvancedCreateView(SuccessMessageMixin, CreateView):
    pass


class AdvancedEditView(SuccessMessageMixin, UpdateView):
    pass


class AdvancedDeleteView(DeleteView):
    """Common confirm view for deleting.

    .. warning ::

        Using this view, objects are deleted permanently after confirming.
        We recommend to include the mixin :class:`reversion.views.RevisionMixin`
        from `django-reversion` to enable soft-delete.
    """

    success_message: Optional[str] = None

    def delete(self, request, *args, **kwargs):
        r = super().delete(request, *args, **kwargs)
        if self.success_message:
            messages.success(self.request, self.success_message)
        return r


class SchoolTermRelatedExtensibleModel(ExtensibleModel):
    """Add relation to school term."""

    objects = CurrentSiteManagerWithoutMigrations.from_queryset(SchoolTermRelatedQuerySet)()

    school_term = models.ForeignKey(
        "core.SchoolTerm",
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("Linked school term"),
        blank=True,
        null=True,
    )

    class Meta:
        abstract = True


class SchoolTermRelatedExtensibleForm(ExtensibleForm):
    """Extensible form for school term related data.

    .. warning::
        This doesn't automatically include the field `school_term` in `fields` or `layout`,
        it just sets an initial value.
    """

    def __init__(self, *args, **kwargs):
        from aleksis.core.models import SchoolTerm  # noqa

        if "instance" not in kwargs:
            kwargs["initial"] = {"school_term": SchoolTerm.current}

        super().__init__(*args, **kwargs)


class PublicFilePreferenceMixin(FilePreference):
    """Uploads a file to the public namespace."""

    upload_path = "public"

    def get_upload_path(self):
        return os.path.join(
            self.upload_path, preferences_settings.FILE_PREFERENCE_UPLOAD_DIR, self.identifier()
        )



from typing import Sequence

from django.contrib.auth.models import Group as DjangoGroup
from django.contrib.auth.models import Permission, User
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.utils.translation import gettext as _

from django_filters import CharFilter, FilterSet, ModelChoiceFilter, ModelMultipleChoiceFilter
from django_select2.forms import ModelSelect2Widget
from guardian.models import GroupObjectPermission, UserObjectPermission
from material import Layout, Row

from aleksis.core.models import Group, GroupType, Person, SchoolTerm


class MultipleCharFilter(CharFilter):
    """Filter for filtering multiple fields with one input.

    >>> multiple_filter = MultipleCharFilter(["name__icontains", "short_name__icontains"])
    """

    def filter(self, qs, value):  # noqa
        q = None
        for field in self.fields:
            if not q:
                q = Q(**{field: value})
            else:
                q = q | Q(**{field: value})
        return qs.filter(q)

    def __init__(self, fields: Sequence[str], *args, **kwargs):
        self.fields = fields
        super().__init__(self, *args, **kwargs)


class GroupFilter(FilterSet):
    school_term = ModelChoiceFilter(queryset=SchoolTerm.objects.all())
    group_type = ModelChoiceFilter(queryset=GroupType.objects.all())
    parent_groups = ModelMultipleChoiceFilter(queryset=Group.objects.all())

    search = MultipleCharFilter(["name__icontains", "short_name__icontains"], label=_("Search"))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.layout = Layout(Row("search"), Row("school_term", "group_type", "parent_groups"))
        self.form.initial = {"school_term": SchoolTerm.current}


class PersonFilter(FilterSet):
    name = MultipleCharFilter(
        [
            "first_name__icontains",
            "additional_name__icontains",
            "last_name__icontains",
            "short_name__icontains",
            "user__username__icontains",
        ],
        label=_("Search by name"),
    )
    contact = MultipleCharFilter(
        [
            "street__icontains",
            "housenumber__icontains",
            "postal_code__icontains",
            "place__icontains",
            "phone_number__icontains",
            "mobile_number__icontains",
            "email__icontains",
        ],
        label=_("Search by contact details"),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.layout = Layout(Row("name", "contact"), Row("sex", "primary_group"))

    class Meta:
        model = Person
        fields = ["sex", "primary_group"]


class PermissionFilter(FilterSet):
    """Common filter for permissions."""

    permission = ModelChoiceFilter(
        queryset=Permission.objects.all(),
        widget=ModelSelect2Widget(
            search_fields=["name__icontains", "codename__icontains"],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
        label=_("Permission"),
    )
    permission__content_type = ModelChoiceFilter(
        queryset=ContentType.objects.all(),
        widget=ModelSelect2Widget(
            search_fields=["app_label__icontains", "model__icontains"],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
        label=_("Content type"),
    )


class UserPermissionFilter(PermissionFilter):
    """Common filter for user permissions."""

    user = ModelChoiceFilter(
        queryset=User.objects.all(),
        widget=ModelSelect2Widget(
            search_fields=["username__icontains", "first_name__icontains", "last_name__icontains"],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
        label=_("User"),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.layout = Layout(Row("user", "permission", "permission__content_type"))

    class Meta:
        fields = ["user", "permission", "permission__content_type"]


class GroupPermissionFilter(PermissionFilter):
    """Common filter for group permissions."""

    group = ModelChoiceFilter(
        queryset=DjangoGroup.objects.all(),
        widget=ModelSelect2Widget(
            search_fields=[
                "name__icontains",
            ],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
        label=_("Group"),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.layout = Layout(Row("group", "permission", "permission__content_type"))

    class Meta:
        fields = ["group", "permission", "permission__content_type"]


class UserGlobalPermissionFilter(UserPermissionFilter):
    """Filter for global user permissions."""

    class Meta(UserPermissionFilter.Meta):
        model = User.user_permissions.through


class GroupGlobalPermissionFilter(GroupPermissionFilter):
    """Filter for global group permissions."""

    class Meta(GroupPermissionFilter.Meta):
        model = DjangoGroup.permissions.through


class UserObjectPermissionFilter(UserPermissionFilter):
    """Filter for object user permissions."""

    class Meta(UserPermissionFilter.Meta):
        model = UserObjectPermission


class GroupObjectPermissionFilter(GroupPermissionFilter):
    """Filter for object group permissions."""

    class Meta(GroupPermissionFilter.Meta):
        model = GroupObjectPermission



from typing import Optional

from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import User
from django.db.models import Model
from django.http import HttpRequest

from django_otp import user_has_device
from guardian.backends import ObjectPermissionBackend
from guardian.shortcuts import get_objects_for_user
from rules import predicate

from ..mixins import ExtensibleModel
from ..models import Group
from .core_helpers import get_content_type_by_perm, get_site_preferences
from .core_helpers import has_person as has_person_helper
from .core_helpers import queryset_rules_filter


def permission_validator(request: HttpRequest, perm: str) -> bool:
    """Check whether the request user has a permission."""
    if request.user:
        return request.user.has_perm(perm)
    return False


def check_global_permission(user: User, perm: str) -> bool:
    """Check whether a user has a global permission."""
    return ModelBackend().has_perm(user, perm)


def check_object_permission(
    user: User, perm: str, obj: Model, checker_obj: Optional[ExtensibleModel] = None
) -> bool:
    """Check whether a user has a permission on an object.

    You can provide a custom ``ObjectPermissionChecker`` for prefetching object permissions
    by annotating an extensible model with ``set_object_permission_checker``.
    This can be the provided object (``obj``)  or a special object
    which is only used to get the checker class (``checker_obj``).
    """
    if not checker_obj:
        checker_obj = obj
    if hasattr(checker_obj, "_permission_checker"):
        return checker_obj._permission_checker.has_perm(perm, obj)
    return ObjectPermissionBackend().has_perm(user, perm, obj)


def has_global_perm(perm: str):
    """Build predicate which checks whether a user has a global permission."""
    name = f"has_global_perm:{perm}"

    @predicate(name)
    def fn(user: User) -> bool:
        return check_global_permission(user, perm)

    return fn


def has_object_perm(perm: str):
    """Build predicate which checks whether a user has a permission on a object."""
    name = f"has_global_perm:{perm}"

    @predicate(name)
    def fn(user: User, obj: Model) -> bool:
        if not obj:
            return False
        return check_object_permission(user, perm, obj)

    return fn


def has_any_object(perm: str, klass):
    """Check if has any object.

    Build predicate which checks whether a user has access
    to objects with the provided permission or rule.
    Differentiates between object-related permissions and rules.
    """
    name = f"has_any_object:{perm}"

    @predicate(name)
    def fn(user: User) -> bool:
        ct_perm = get_content_type_by_perm(perm)
        # In case an object-related permission with the same ContentType class as the given class
        # is passed, the optimized django-guardian get_objects_for_user function is used.
        if ct_perm and ct_perm.model_class() == klass:
            return get_objects_for_user(user, perm, klass).exists()
        # In other cases, it is checked for each object of the given model whether the current user
        # fulfills the given rule.
        else:
            return queryset_rules_filter(user, klass.objects.all(), perm).exists()

    return fn


def is_site_preference_set(section: str, pref: str):
    """Check the boolean value of a given site preference."""
    name = f"check_site_preference:{section}__{pref}"

    @predicate(name)
    def fn() -> bool:
        return bool(get_site_preferences()[f"{section}__{pref}"])

    return fn


@predicate
def has_person(user: User) -> bool:
    """Predicate which checks whether a user has a linked person."""
    return has_person_helper(user)


@predicate
def is_current_person(user: User, obj: Model) -> bool:
    """Predicate which checks if the provided object is the person linked to the user object."""
    return user.person == obj


@predicate
def is_group_owner(user: User, group: Group) -> bool:
    """Predicate which checks if the user is a owner of the provided group."""
    return user.person in group.owners.all()


@predicate
def is_group_member(user: User, group: Group) -> bool:
    """Predicate which checks if the user is a member of the provided group."""
    return user.person in group.members.all()


@predicate
def is_notification_recipient(user: User, obj: Model) -> bool:
    """Check if is a notification recipient.

    Predicate which checks whether the recipient of the
    notification a user wants to mark read is this user.
    """
    return user == obj.recipient.user


def contains_site_preference_value(section: str, pref: str, value: str):
    """Check if given site preference contains a value."""
    name = f"check_site_preference_value:{section}__{pref}"

    @predicate(name)
    def fn() -> bool:
        return bool(value in get_site_preferences()[f"{section}__{pref}"])

    return fn


@predicate
def has_activated_2fa(user: User) -> bool:
    """Check if the user has activated two-factor authentication."""
    return user_has_device(user)


@predicate
def is_assigned_to_current_person(user: User, obj: Model) -> bool:
    """Check if the object is assigned to the current person."""
    return getattr(obj, "person", None) == user.person



import random
import string

from django import template
from django.shortcuts import reverse

from bs4 import BeautifulSoup

register = template.Library()


@register.filter
def add_class_to_el(value: str, arg: str) -> str:
    """Add a CSS class to every occurence of an element type.

    :Example:

    .. code-block::

        {{ mymodel.myhtmlfield|add_class_to_el:"ul,browser-default" }}
    """
    el, cls = arg.split(",")
    soup = BeautifulSoup(value, "html.parser")

    for el in soup.find_all(el):
        el["class"] = el.get("class", []) + [cls]

    return str(soup)


@register.simple_tag
def generate_random_id(prefix: str, length: int = 10) -> str:
    """Generate a random ID for templates.

    :Example:

    .. code-block::

        {% generate_random_id "prefix-" %}
    """
    return prefix + "".join(
        random.choice(string.ascii_lowercase) for i in range(length)  # noqa: S311
    )


@register.simple_tag(takes_context=True)
def absolute_url(context, view_name, *args, **kwargs):
    request = context["request"]
    return request.build_absolute_uri(reverse(view_name, args=args, kwargs=kwargs))
